#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/07/2004
#
#------------------------------------------------------------------------------

""" Defines the UI class used to represent an active traits-based user
    interface.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

import shelve
import os

from traits.api import (
    Any,
    Bool,
    Callable,
    DictStrAny,
    Event,
    HasPrivateTraits,
    Instance,
    Int,
    List,
    Property,
    Str,
    TraitError,
    on_trait_change,
    property_depends_on)

from traits.trait_base import traits_home, is_str

from .editor import Editor

from .view_elements import ViewElements

from .handler import Handler, ViewHandler

from .toolkit import toolkit

from .ui_info import UIInfo

from .item import Item

from .group import Group, ShadowGroup

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# List of **kind** types for views that must have a **parent** window specified
kind_must_have_parent = ('panel', 'subpanel')

#-------------------------------------------------------------------------
#  'UI' class:
#-------------------------------------------------------------------------


class UI(HasPrivateTraits):
    """ Information about the user interface for a View.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The ViewElements object from which this UI resolves Include items
    view_elements = Instance(ViewElements)

    # Context objects that the UI is editing
    context = DictStrAny

    # Handler object used for event handling
    handler = Instance(Handler)

    # View template used to construct the user interface
    view = Instance('traitsui.view.View')

    # Panel or dialog associated with the user interface
    control = Any

    # The parent UI (if any) of this UI
    parent = Instance('UI')

    # Toolkit-specific object that "owns" **control**
    owner = Any

    # UIInfo object containing context or editor objects
    info = Instance(UIInfo)

    # Result from a modal or wizard dialog:
    result = Bool(False)

    # Undo and Redo history
    history = Any

    # The KeyBindings object (if any) for this UI:
    key_bindings = Property  # Instance( KeyBindings )

    # The unique ID for this UI for persistence
    id = Str

    # Have any modifications been made to UI contents?
    modified = Bool(False)

    # Event when the user interface has changed
    updated = Event(Bool)

    # Title of the dialog, if any
    title = Str

    # The ImageResource of the icon, if any
    icon = Any

    # Should the created UI have scroll bars?
    scrollable = Bool(False)

    # The number of currently pending editor error conditions
    errors = Int

    # The code used to rebuild an updated user interface
    rebuild = Callable

    # Set to True when the UI has finished being destroyed.
    destroyed = Bool(False)

    #-- Private Traits -------------------------------------------------------

    # Original context when used with a modal dialog
    _context = DictStrAny

    # Copy of original context used for reverting changes
    _revert = DictStrAny

    # List of methods to call once the user interface is created
    _defined = List

    # List of (visible_when,Editor) pairs
    _visible = List

    # List of (enabled_when,Editor) pairs
    _enabled = List

    # List of (checked_when,Editor) pairs
    _checked = List

    # Search stack used while building a user interface
    _search = List

    # List of dispatchable Handler methods
    _dispatchers = List

    # List of editors used to build the user interface
    _editors = List

    # List of names bound to the **info** object
    _names = List

    # Index of currently the active group in the user interface
    _active_group = Int

    # List of top-level groups used to build the user interface
    _groups = Property
    _groups_cache = Any

    # Count of levels of nesting for undoable actions
    _undoable = Int(-1)

    # Code used to rebuild an updated user interface
    _rebuild = Callable

    # The statusbar listeners that have been set up:
    _statusbar = List

    # Does the UI contain any scrollable widgets?
    #
    # The _scrollable trait is set correctly, but not used currently because
    # its value is arrived at too late to be of use in building the UI.
    _scrollable = Bool(False)

    # List of traits that are reset when a user interface is recycled
    # (i.e. rebuilt).
    recyclable_traits = [
        '_context', '_revert', '_defined', '_visible', '_enabled', '_checked',
        '_search', '_dispatchers', '_editors', '_names', '_active_group',
        '_undoable', '_rebuild', '_groups_cache'
    ]

    # List of additional traits that are discarded when a user interface is
    # disposed.
    disposable_traits = [
        'view_elements', 'info', 'handler', 'context', 'view', 'history',
        'key_bindings', 'icon', 'rebuild',
    ]

    #-------------------------------------------------------------------------
    #  Initializes the traits object:
    #-------------------------------------------------------------------------

    def traits_init(self):
        """ Initializes the traits object.
        """
        self.info = UIInfo(ui=self)
        self.handler.init_info(self.info)

    #-------------------------------------------------------------------------
    #  Creates a user interface from the associated View template object:
    #-------------------------------------------------------------------------

    def ui(self, parent, kind):
        """ Creates a user interface from the associated View template object.
        """
        if (parent is None) and (kind in kind_must_have_parent):
            kind = 'live'
        self.view.on_trait_change(self._updated_changed, 'updated',
                                  dispatch='ui')
        self.rebuild = getattr(toolkit(), 'ui_' + kind)
        self.rebuild(self, parent)

    #-------------------------------------------------------------------------
    #  Disposes of the contents of a user interface:
    #-------------------------------------------------------------------------

    def dispose(self, result=None, abort=False):
        """ Disposes of the contents of a user interface.
        """
        if result is not None:
            self.result = result

        # Only continue if the view has not already been disposed of:
        if self.control is not None:
            # Save the user preference information for the user interface:
            if not abort:
                self.save_prefs()

            # Finish disposing of the user interface:
            self.finish()

    #-------------------------------------------------------------------------
    #  Recycles the user interface prior to rebuilding it:
    #-------------------------------------------------------------------------

    def recycle(self):
        """ Recycles the user interface prior to rebuilding it.
        """
        # Reset all user interface editors:
        self.reset(destroy=False)

        # Discard any context object associated with the ui view control:
        self.control._object = None

        # Reset all recyclable traits:
        self.reset_traits(self.recyclable_traits)

    #-------------------------------------------------------------------------
    #  Finishes a user interface:
    #-------------------------------------------------------------------------

    def finish(self):
        """ Finishes disposing of a user interface.
        """

        # Reset the contents of the user interface
        self.reset(destroy=False)

        # Make sure that 'visible', 'enabled', and 'checked' handlers are not
        # called after the editor has been disposed:
        for object in self.context.values():
            object.on_trait_change(self._evaluate_when, remove=True)

        # Notify the handler that the view has been closed:
        self.handler.closed(self.info, self.result)

        # Clear the back-link from the UIInfo object to us:
        self.info.ui = None

        # Destroy the view control:
        self.control._object = None
        toolkit().destroy_control(self.control)
        self.control = None

        # Dispose of any KeyBindings object we reference:
        if self.key_bindings is not None:
            self.key_bindings.dispose()

        # Break the linkage to any objects in the context dictionary:
        self.context.clear()

        # Remove specified symbols from our dictionary to aid in clean-up:
        self.reset_traits(self.recyclable_traits)
        self.reset_traits(self.disposable_traits)

        self.destroyed = True

    #-------------------------------------------------------------------------
    #  Resets the contents of the user interface:
    #-------------------------------------------------------------------------

    def reset(self, destroy=True):
        """ Resets the contents of a user interface.
        """
        for editor in self._editors:
            if editor._ui is not None:
                # Propagate result to enclosed ui objects:
                editor._ui.result = self.result
            editor.dispose()

            # Zap the control. If there are pending events for the control in
            # the UI queue, the editor's '_update_editor' method will see that
            # the control is None and discard the update request:
            editor.control = None

        # Remove any statusbar listeners that have been set up:
        for object, handler, name in self._statusbar:
            object.on_trait_change(handler, name, remove=True)

        del self._statusbar[:]

        if destroy:
            toolkit().destroy_children(self.control)

        for dispatcher in self._dispatchers:
            dispatcher.remove()

    #-------------------------------------------------------------------------
    #  Find the definition of the specified Include object in the current user
    #  interface building context:
    #-------------------------------------------------------------------------

    def find(self, include):
        """ Finds the definition of the specified Include object in the current
            user interface building context.
        """
        context = self.context
        result = None

        # Get the context 'object' (if available):
        if len(context) == 1:
            object = context.values()[0]
        else:
            object = context.get('object')

        # Try to use our ViewElements objects:
        ve = self.view_elements

        # If none specified, try to get it from the UI context:
        if (ve is None) and (object is not None):
            # Use the context object's ViewElements (if available):
            ve = object.trait_view_elements()

        # Ask the ViewElements to find the requested item for us:
        if ve is not None:
            result = ve.find(include.id, self._search)

        # If not found, then try to search the 'handler' and 'object' for a
        # method we can call that will define it:
        if result is None:
            handler = context.get('handler')
            if handler is not None:
                method = getattr(handler, include.id, None)
                if callable(method):
                    result = method()

            if (result is None) and (object is not None):
                method = getattr(object, include.id, None)
                if callable(method):
                    result = method()

        return result

    #-------------------------------------------------------------------------
    #  Returns the current search stack level:
    #-------------------------------------------------------------------------

    def push_level(self):
        """ Returns the current search stack level.
        """
        return len(self._search)

    #-------------------------------------------------------------------------
    #  Restores a previously pushed search stack level:
    #-------------------------------------------------------------------------

    def pop_level(self, level):
        """ Restores a previously pushed search stack level.
        """
        del self._search[: len(self._search) - level]

    #-------------------------------------------------------------------------
    #  Performs all post user interface creation processing:
    #-------------------------------------------------------------------------

    def prepare_ui(self):
        """ Performs all processing that occurs after the user interface is
            created.
        """
        # Invoke all of the editor 'name_defined' methods we've accumulated:
        info = self.info.set(initialized=False)
        for method in self._defined:
            method(info)

        # Then reset the list, since we don't need it anymore:
        del self._defined[:]

        # Synchronize all context traits with associated editor traits:
        self.sync_view()

        # Hook all keyboard events:
        toolkit().hook_events(self, self.control, 'keys', self.key_handler)

        # Hook all events if the handler is an extended 'ViewHandler':
        handler = self.handler
        if isinstance(handler, ViewHandler):
            toolkit().hook_events(self, self.control)

        # Invoke the handler's 'init' method, and abort if it indicates
        # failure:
        if handler.init(info) == False:
            raise TraitError('User interface creation aborted')

        # For each Handler method whose name is of the form
        # 'object_name_changed', where 'object' is the name of an object in the
        # UI's 'context', create a trait notification handler that will call
        # the method whenever 'object's 'name' trait changes. Also invoke the
        # method immediately so initial user interface state can be correctly
        # set:
        context = self.context
        for name in self._each_trait_method(handler):
            if name[-8:] == '_changed':
                prefix = name[:-8]
                col = prefix.find('_', 1)
                if col >= 0:
                    object = context.get(prefix[: col])
                    if object is not None:
                        method = getattr(handler, name)
                        trait_name = prefix[col + 1:]
                        self._dispatchers.append(Dispatcher(
                            method, info, object, trait_name))
                        if object.base_trait(trait_name).type != 'event':
                            method(info)

        # If there are any Editor object's whose 'visible', 'enabled' or
        # 'checked' state is controlled by a 'visible_when', 'enabled_when' or
        # 'checked_when' expression, set up an 'anytrait' changed notification
        # handler on each object in the 'context' that will cause the 'visible',
        # 'enabled' or 'checked' state of each affected Editor to be set. Also
        # trigger the evaluation immediately, so the visible, enabled or checked
        # state of each Editor can be correctly initialized:
        if (len(self._visible) +
            len(self._enabled) +
                len(self._checked)) > 0:
            for object in context.values():
                object.on_trait_change(self._evaluate_when, dispatch='ui')
            self._do_evaluate_when(at_init=True)

        # Indicate that the user interface has been initialized:
        info.initialized = True

    #-------------------------------------------------------------------------
    #  Synchronize context object traits with view editor traits:
    #-------------------------------------------------------------------------

    def sync_view(self):
        """ Synchronize context object traits with view editor traits.
        """
        for name, object in self.context.items():
            self._sync_view(name, object, 'sync_to_view', 'from')
            self._sync_view(name, object, 'sync_from_view', 'to')
            self._sync_view(name, object, 'sync_with_view', 'both')

    def _sync_view(self, name, object, metadata, direction):
        info = self.info
        for trait_name, trait in object.traits(**{metadata: is_str}).items():
            for sync in getattr(trait, metadata).split(','):
                try:
                    editor_id, editor_name = [item.strip()
                                              for item in sync.split('.')]
                except:
                    raise TraitError(
                        "The '%s' metadata for the '%s' trait in "
                        "the '%s' context object should be of the form: "
                        "'id1.trait1[,...,idn.traitn]." %
                        (metadata, trait_name, name))

                editor = getattr(info, editor_id, None)
                if editor is not None:
                    editor.sync_value('%s.%s' % (name, trait_name),
                                      editor_name, direction)
                else:
                    raise TraitError(
                        "No editor with id = '%s' was found for "
                        "the '%s' metadata for the '%s' trait in the '%s' "
                        "context object." %
                        (editor_id, metadata, trait_name, name))

    #-------------------------------------------------------------------------
    #  Gets the current value of a specified extended trait name:
    #-------------------------------------------------------------------------

    def get_extended_value(self, name):
        """ Gets the current value of a specified extended trait name.
        """
        names = name.split('.')
        if len(names) > 1:
            value = self.context[names[0]]
            del names[0]
        else:
            value = self.context['object']

        for name in names:
            value = getattr(value, name)

        return value

    #-------------------------------------------------------------------------
    #  Restores any saved user preference information associated with the UI:
    #-------------------------------------------------------------------------

    def restore_prefs(self):
        """ Retrieves and restores any saved user preference information
        associated with the UI.
        """
        id = self.id
        if id != '':
            db = self.get_ui_db()
            if db is not None:
                try:
                    ui_prefs = db.get(id)
                    db.close()
                    return self.set_prefs(ui_prefs)
                except:
                    pass

        return None

    #-------------------------------------------------------------------------
    #  Restores user preference information for the UI:
    #-------------------------------------------------------------------------

    def set_prefs(self, prefs):
        """ Sets the values of user preferences for the UI.
        """
        if isinstance(prefs, dict):
            info = self.info
            for name in self._names:
                editor = getattr(info, name, None)
                if isinstance(editor, Editor) and (editor.ui is self):
                    editor_prefs = prefs.get(name)
                    if editor_prefs is not None:
                        editor.restore_prefs(editor_prefs)

            if self.key_bindings is not None:
                key_bindings = prefs.get('$')
                if key_bindings is not None:
                    self.key_bindings.merge(key_bindings)

            return prefs.get('')

        return None

    #-------------------------------------------------------------------------
    #  Saves any user preference information associated with the UI:
    #-------------------------------------------------------------------------

    def save_prefs(self, prefs=None):
        """ Saves any user preference information associated with the UI.
        """
        if prefs is None:
            toolkit().save_window(self)
            return

        id = self.id
        if id != '':
            db = self.get_ui_db(mode='c')
            if db is not None:
                db[id] = self.get_prefs(prefs)
                db.close()

    #-------------------------------------------------------------------------
    #  Gets the preferences to be saved for the user interface:
    #-------------------------------------------------------------------------

    def get_prefs(self, prefs=None):
        """ Gets the preferences to be saved for the user interface.
        """
        ui_prefs = {}
        if prefs is not None:
            ui_prefs[''] = prefs

        if self.key_bindings is not None:
            ui_prefs['$'] = self.key_bindings

        info = self.info
        for name in self._names:
            editor = getattr(info, name, None)
            if isinstance(editor, Editor) and (editor.ui is self):
                prefs = editor.save_prefs()
                if prefs is not None:
                    ui_prefs[name] = prefs

        return ui_prefs

    #-------------------------------------------------------------------------
    #  Gets a reference to the traits UI preference database:
    #-------------------------------------------------------------------------

    def get_ui_db(self, mode='r'):
        """ Returns a reference to the Traits UI preference database.
        """
        try:
            return shelve.open(os.path.join(traits_home(), 'traits_ui'),
                               flag=mode, protocol=-1)
        except:
            return None

    #-------------------------------------------------------------------------
    #  Returns a list of editors for the given trait name.
    #-------------------------------------------------------------------------

    def get_editors(self, name):
        """ Returns a list of editors for the given trait name.
        """
        return [editor for editor in self._editors if editor.name == name]

    #-------------------------------------------------------------------------
    #  Returns the list of editor error controls contained by the user
    #  interface:
    #-------------------------------------------------------------------------

    def get_error_controls(self):
        """ Returns the list of editor error controls contained by the user
            interface.
        """
        controls = []
        for editor in self._editors:
            control = editor.get_error_control()
            if isinstance(control, list):
                controls.extend(control)
            else:
                controls.append(control)

        return controls

    #-------------------------------------------------------------------------
    #  Adds a Handler method to the list of methods to be called once the user
    #  interface has been constructed:
    #-------------------------------------------------------------------------

    def add_defined(self, method):
        """ Adds a Handler method to the list of methods to be called once the
            user interface has been constructed.
        """
        self._defined.append(method)

    #-------------------------------------------------------------------------
    #  Add's a conditionally enabled Editor object to the list of monitored
    #  'visible_when' objects:
    #-------------------------------------------------------------------------

    def add_visible(self, visible_when, editor):
        """ Adds a conditionally enabled Editor object to the list of monitored
            'visible_when' objects.
        """
        try:
            self._visible.append((compile(visible_when, '<string>', 'eval'),
                                  editor))
        except:
            pass
            # fixme: Log an error here...

    #-------------------------------------------------------------------------
    #  Add's a conditionally enabled Editor object to the list of monitored
    #  'enabled_when' objects:
    #-------------------------------------------------------------------------

    def add_enabled(self, enabled_when, editor):
        """ Adds a conditionally enabled Editor object to the list of monitored
            'enabled_when' objects.
        """
        try:
            self._enabled.append((compile(enabled_when, '<string>', 'eval'),
                                  editor))
        except:
            pass
            # fixme: Log an error here...

    #-------------------------------------------------------------------------
    #  Add's a conditionally checked (menu/toolbar) Editor object to the list of
    #  monitored 'checked_when' objects:
    #-------------------------------------------------------------------------

    def add_checked(self, checked_when, editor):
        """ Adds a conditionally enabled (menu) Editor object to the list of
            monitored 'checked_when' objects.
        """
        try:
            self._checked.append((compile(checked_when, '<string>', 'eval'),
                                  editor))
        except:
            pass
            # fixme: Log an error here...

    #-------------------------------------------------------------------------
    #  Performs an 'undoable' action:
    #-------------------------------------------------------------------------

    def do_undoable(self, action, *args, **kw):
        """ Performs an action that can be undone.
        """
        undoable = self._undoable
        try:
            if (undoable == -1) and (self.history is not None):
                self._undoable = self.history.now

            action(*args, **kw)
        finally:
            if undoable == -1:
                self._undoable = -1

    #-------------------------------------------------------------------------
    #  Routes a 'hooked' event to the correct handler method:
    #-------------------------------------------------------------------------

    def route_event(self, event):
        """ Routes a "hooked" event to the correct handler method.
        """
        toolkit().route_event(self, event)

    #-------------------------------------------------------------------------
    #  Handles key events when the view has a set of KeyBindings:
    #-------------------------------------------------------------------------

    def key_handler(self, event, skip=True):
        """ Handles key events.
        """
        key_bindings = self.key_bindings
        handled = ((key_bindings is not None) and
                   key_bindings.do(event, [], self.info,
                                   recursive=(self.parent is None)))

        if (not handled) and (self.parent is not None):
            handled = self.parent.key_handler(event, False)

        if (not handled) and skip:
            toolkit().skip_event(event)

        return handled

    #-------------------------------------------------------------------------
    #  Evaluates a specified function in the UI's context:
    #-------------------------------------------------------------------------

    def evaluate(self, function, *args, **kw_args):
        """ Evaluates a specified function in the UI's **context**.
        """
        if function is None:
            return None

        if callable(function):
            return function(*args, **kw_args)

        context = self.context.copy()
        context['ui'] = self
        context['handler'] = self.handler
        return eval(function, globals(), context)(*args, **kw_args)

    #-------------------------------------------------------------------------
    #  Evaluates an expression in the UI's 'context' and returns the result:
    #-------------------------------------------------------------------------

    def eval_when(self, when, result=True):
        """ Evaluates an expression in the UI's **context** and returns the
            result.
        """
        context = self._get_context(self.context)
        try:
            result = eval(when, globals(), context)
        except:
            from traitsui.api import raise_to_debug
            raise_to_debug()

        del context['ui']

        return result

    #-------------------------------------------------------------------------
    #  Gets the context to use for evaluating an expression:
    #-------------------------------------------------------------------------

    def _get_context(self, context):
        """ Gets the context to use for evaluating an expression.
        """
        name = 'object'
        n = len(context)
        if (n == 2) and ('handler' in context):
            for name, value in context.items():
                if name != 'handler':
                    break
        elif n == 1:
            name = context.keys()[0]

        value = context.get(name)
        if value is not None:
            context2 = value.trait_get()
            context2.update(context)
        else:
            context2 = context.copy()

        context2['ui'] = self

        return context2

    #-------------------------------------------------------------------------
    #  Sets the 'visible', 'enabled' and/or 'checked' state for all Editors
    #  controlled by a 'visible_when', 'enabled_when' or 'checked_when'
    #  expression:
    #-------------------------------------------------------------------------

    def _evaluate_when(self):
        """ Set the 'visible', 'enabled', and 'checked' states for all Editors
            controlled by a 'visible_when', 'enabled_when' or 'checked_when'
            expression.
        """
        self._do_evaluate_when(at_init=False)

    def _do_evaluate_when(self, at_init=False):
        """ Set the 'visible', 'enabled', and 'checked' states for all Editors.

        This function does the job of _evaluate_when. We define it here to
        work around the traits dispatching mechanism that automatically
        determines the number of arguments of a notification method.

        :attr:`at_init` is set to true when this function is called the first
        time at initialization. In that case, we want to force the state of
        the items to be set (normally it is set only if it changes).
        """
        self._evaluate_condition(self._visible, 'visible', at_init)
        self._evaluate_condition(self._enabled, 'enabled', at_init)
        self._evaluate_condition(self._checked, 'checked', at_init)

    #-------------------------------------------------------------------------
    #  Evaluates a list of ( eval, editor ) pairs and sets a specified trait on
    #  each editor to reflect the boolean truth of the expression evaluated:
    #-------------------------------------------------------------------------

    def _evaluate_condition(self, conditions, trait, at_init=False):
        """ Evaluates a list of (eval, editor) pairs and sets a specified trait
        on each editor to reflect the Boolean value of the expression.

        1) All conditions are evaluated
        2) The elements whose condition evaluates to False are updated
        3) The elements whose condition evaluates to True are updated

        E.g., we first make invisible all elements for which 'visible_when'
        evaluates to False, and then we make visible the ones
        for which 'visible_when' is True. This avoids mutually exclusive
        elements to be visible at the same time, and thus making a dialog
        unnecessarily large.

        The state of an editor is updated only when it changes, unless
        at_init is set to True.

        Parameters
        ----------
        conditions : list of (str, Editor) tuple
            A list of tuples, each formed by 1) a string that contains a
            condition that evaluates to either True or False, and
            2) the editor whose state depends on the condition

        trait : str
            The trait that is set by the condition.
            Either 'visible, 'enabled', or 'checked'.

        at_init : bool
            If False, the state of an editor is set only when it changes
            (e.g., a visible element would not be updated to visible=True
            again). If True, the state is always updated (used at
            initialization).
        """

        context = self._get_context(self.context)

        # list of elements that should be activated
        activate = []
        # list of elements that should be de-activated
        deactivate = []

        for when, editor in conditions:
            try:
                cond_value = eval(when, globals(), context)
                editor_state = getattr(editor, trait)

                # add to update lists only if at_init is True (called on
                # initialization), or if the editor state has to change

                if cond_value and (at_init or not editor_state):
                    activate.append(editor)

                if not cond_value and (at_init or editor_state):
                    deactivate.append(editor)

            except Exception:
                # catch errors in the validate_when expression
                from traitsui.api import raise_to_debug
                raise_to_debug()

        # update the state of the editors
        for editor in deactivate:
            setattr(editor, trait, False)
        for editor in activate:
            setattr(editor, trait, True)

    #-------------------------------------------------------------------------
    #  Implementation of the '_groups' property:
    #  (Returns the top-level Groups for the view (after resolving Includes))
    #-------------------------------------------------------------------------

    def _get__groups(self):
        """ Returns the top-level Groups for the view (after resolving
        Includes. (Implements the **_groups** property.)
        """
        if self._groups_cache is None:
            shadow_group = self.view.content.get_shadow(self)
            self._groups_cache = shadow_group.get_content()
            for item in self._groups_cache:
                if isinstance(item, Item):
                    self._groups_cache = [
                        ShadowGroup(shadow=Group(*self._groups_cache),
                                    content=self._groups_cache,
                                    groups=1)
                    ]
                    break
        return self._groups_cache

    #-- Property Implementations ---------------------------------------------

    @property_depends_on('view, context')
    def _get_key_bindings(self):
        view, context = self.view, self.context
        if (view is None) or (context is None):
            return None

        # Get the KeyBindings object to use:
        values = context.values()
        key_bindings = view.key_bindings
        if key_bindings is None:
            from .key_bindings import KeyBindings

            return KeyBindings(controllers=values)

        return key_bindings.clone(controllers=values)

    #-- Traits Event Handlers ------------------------------------------------

    def _updated_changed(self):
        if self.rebuild is not None:
            toolkit().rebuild_ui(self)

    def _title_changed(self):
        if self.control is not None:
            toolkit().set_title(self)

    def _icon_changed(self):
        if self.control is not None:
            toolkit().set_icon(self)

    @on_trait_change('parent, view, context')
    def _pvc_changed(self):
        parent = self.parent
        if (parent is not None) and (self.key_bindings is not None):
            # If we don't have our own history, use our parent's:
            if self.history is None:
                self.history = parent.history

            # Link our KeyBindings object as a child of our parent's
            # KeyBindings object (if any):
            if parent.key_bindings is not None:
                parent.key_bindings.children.append(self.key_bindings)

#-------------------------------------------------------------------------
#  'Dispatcher' class:
#-------------------------------------------------------------------------


class Dispatcher(object):

    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, method, info, object, method_name):
        """ Initializes the object.
        """
        self.method = method
        self.info = info
        self.object = object
        self.method_name = method_name
        object.on_trait_change(self.dispatch, method_name, dispatch='ui')

    #-------------------------------------------------------------------------
    #  Dispatches the method:
    #-------------------------------------------------------------------------

    def dispatch(self):
        """ Dispatches the method.
        """
        self.method(self.info)

    #-------------------------------------------------------------------------
    #  Remove the dispatcher:
    #-------------------------------------------------------------------------

    def remove(self):
        """ Removes the dispatcher.
        """
        self.object.on_trait_change(self.dispatch, self.method_name,
                                    remove=True)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Judah De Paula
#  Date:   10/7/2008
#
#------------------------------------------------------------------------------

""" A Traits UI editor that wraps a WX calendar panel.
"""

from __future__ import absolute_import

from traits.trait_types import Bool, Int, Enum, Str
from ..ui_traits import AView
from ..editor_factory import EditorFactory


class DateEditor(EditorFactory):
    """ Editor factory for date/time editors.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    #-- ReadonlyEditor traits ------------------------------------------------

    # Message to show when Date is None.
    message = Str('Undefined')

    # The string representation of the date to show.  Uses time.strftime
    # format.
    strftime = Str('%B %d %Y (%a)')

    # An optional view to display when a read-only text editor is clicked:
    view = AView

    #-- CustomEditor traits --------------------------------------------------

    # Should users be able to pick future dates when using the CustomEditor?
    allow_future = Bool(True)

    # How many months to show at a time.
    months = Int(3)

    # True: Must be a List of Dates.  False: Must be a Date instance.
    multi_select = Bool(False)

    # When a user multi-selects entries and some of those entries are already
    # selected and some are not, what should be the behavior for the seletion?
    # Options::
    #
    #     'toggle'     -- Toggle each day to the opposite of the current state.
    #     'on'         -- Always turn them on.
    #     'off'        -- Always turn them off.
    #     'max_change' -- Change all to same state, with most days changing.
    #                     For example 1 selected and 9 not, then they would
    #                     all get selected.
    #     'min_change' -- Change all to same state, with min days changing.
    #                     For example 1 selected and 9 not, then they would
    #                     all get unselected.
    on_mixed_select = Enum('toggle', 'on', 'off', 'max_change', 'min_change')

    # How much space to put between the individual months.
    padding = Int(5)

    # Does the user have to hold down Shift for the left-click multiselect?
    shift_to_select = Bool(False)



#-------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   09/27/2005
#
#-------------------------------------------------------------------------

""" Editor that displays an interactive Python shell.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.shell_editor file.
from traitsui.editors.shell_editor \
    import _ShellEditor as BaseShellEditor, ToolkitEditorFactory

from editor \
    import Editor

#-------------------------------------------------------------------------
#  'ShellEditor' class:
#-------------------------------------------------------------------------


class _ShellEditor(BaseShellEditor, Editor):
    """ Editor that displays an interactive Python shell.
    """
    pass



"""
fs.wrapfs.lazyfs
================

A class for lazy initialization of an FS object.

This module provides the class LazyFS, an FS wrapper class that can lazily
initialize its underlying FS object.

"""

import sys

try:
    from threading import Lock
except ImportError:
    from fs.base import DummyLock as Lock

from fs.base import FS
from fs.wrapfs import WrapFS


class LazyFS(WrapFS):
    """Simple 'lazy initialization' for FS objects.

    This FS wrapper can be created with an FS instance, an FS class, or a
    (class,args,kwds) tuple.  The actual FS instance will be created on demand
    the first time it is accessed.
    """

    def __init__(self, fs):
        super(LazyFS, self).__init__(fs)
        self._lazy_creation_lock = Lock()

    def __unicode__(self):
        try:
            wrapped_fs = self.__dict__["wrapped_fs"]
        except KeyError:
            #  It appears that python2.5 has trouble printing out
            #  classes that define a __unicode__ method.
            try:
                return u"<LazyFS: %s>" % (self._fsclass,)
            except TypeError:
                try:
                    return u"<LazyFS: %s>" % (self._fsclass.__name__,)
                except AttributeError:
                    return u"<LazyFS: <unprintable>>"
        else:
            return u"<LazyFS: %s>" % (wrapped_fs,)

    def __getstate__(self):
        state = super(LazyFS,self).__getstate__()
        del state["_lazy_creation_lock"]
        return state

    def __setstate__(self, state):
        super(LazyFS,self).__setstate__(state)
        self._lazy_creation_lock = Lock()

    def _get_wrapped_fs(self):
        """Obtain the wrapped FS instance, creating it if necessary."""
        try:
            fs = self.__dict__["wrapped_fs"]
        except KeyError:
            self._lazy_creation_lock.acquire()
            try:
                try:
                    fs = self.__dict__["wrapped_fs"]
                except KeyError:
                    fs = self._fsclass(*self._fsargs,**self._fskwds)
                    self.__dict__["wrapped_fs"] = fs
            finally:
                self._lazy_creation_lock.release()
        return fs

    def _set_wrapped_fs(self, fs):
        if isinstance(fs,FS):
            self.__dict__["wrapped_fs"] = fs
        elif isinstance(fs,type):
            self._fsclass = fs
            self._fsargs = []
            self._fskwds = {}
        elif fs is None:
            del self.__dict__['wrapped_fs']
        else:
            self._fsclass = fs[0]
            try:
                self._fsargs = fs[1]
            except IndexError:
                self._fsargs = []
            try:
                self._fskwds = fs[2]
            except IndexError:
                self._fskwds = {}

    wrapped_fs = property(_get_wrapped_fs,_set_wrapped_fs)

    def setcontents(self, path, data, chunk_size=64*1024):
        return self.wrapped_fs.setcontents(path, data, chunk_size=chunk_size)

    def close(self):
        if not self.closed:
            #  If it was never initialized, create a fake one to close.
            if "wrapped_fs" not in self.__dict__:
                self.__dict__["wrapped_fs"] = FS()
            super(LazyFS,self).close()





#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" Base class for all tree label providers. """


# Local imports.
from .label_provider import LabelProvider


class TreeLabelProvider(LabelProvider):
    """ Base class for all tree label providers.

    By default an element has no label image, and 'str' is used to generate its
    label text.

    """

    ###########################################################################
    # 'LabelProvider' interface.
    ###########################################################################

    def set_text(self, viewer, element, text):
        """ Sets the text representation of a node.

        Returns True if setting the text succeeded, otherwise False.

        """

        return len(text.strip()) > 0

    ###########################################################################
    # 'TreeLabelProvider' interface.
    ###########################################################################

    def get_drag_value(self, viewer, element):
        """ Get the value that is dragged for an element.

        By default the drag value is the element itself.

        """

        return element

    def is_collapsible(self, viewer, element):
        """ Returns True is the element is collapsible, otherwise False. """

        return True

    def is_expandable(self, viewer, node):
        """ Returns True is the node is expandanble, otherwise False. """

        return True

#### EOF ######################################################################



#!/usr/bin/env python

"""
test code for kmz files
"""

from post_gnome import kmz_particles


def test_init():
    kmz_particles.Writer('junk')


def test_copy():
    name = kmz_particles.nc2kmz('sample.nc')
    assert name == 'sample.kmz'

    ## fixme -- should put some real tests in here!
    assert True

# def test_copy_real_gnome():
#     """
#     see if converting to kml works for a real GNOME file
#     """
#     name = kmz_particles.nc2kmz('chesapeake_bay_example.nc')
#     assert name == 'chesapeake_bay_example.kmz'

#     ## fixme -- should put some real tests in here!
#     assert False




#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" A page in a wizard. """


# Major package imports.
import wx

# Enthought library imports.
from traits.api import Bool, HasTraits, provides, Str, Tuple, Unicode
from pyface.api import HeadingText
from pyface.wizard.i_wizard_page import IWizardPage, MWizardPage


@provides(IWizardPage)
class WizardPage(MWizardPage, HasTraits):
    """ The toolkit specific implementation of a WizardPage.

    See the IWizardPage interface for the API documentation.

    """


    #### 'IWizardPage' interface ##############################################

    id = Str

    next_id = Str

    last_page = Bool(False)

    complete = Bool(False)

    heading = Unicode

    subheading = Unicode

    size = Tuple

    ###########################################################################
    # 'IWizardPage' interface.
    ###########################################################################

    def create_page(self, parent):
        """ Creates the wizard page. """

        # FIXME: implement support for the size trait.

        panel = wx.Panel(parent, -1, style=wx.CLIP_CHILDREN)
        sizer = wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)
        panel.SetAutoLayout(True)

        # The 'pretty' heading ;^)
        if len(self.heading) > 0:
            title = HeadingText(panel, text=self.heading)
            sizer.Add(title.control, 0, wx.EXPAND | wx.BOTTOM, 5)

        if len(self.subheading) > 0:
            subtitle = wx.StaticText(panel, -1, self.subheading)
            sizer.Add(subtitle, 0, wx.EXPAND | wx.BOTTOM, 5)

        # The page content.
        content = self._create_page_content(panel)
        sizer.Add(content, 1, wx.EXPAND)

        return panel

    ###########################################################################
    # Protected 'IWizardPage' interface.
    ###########################################################################

    def _create_page_content(self, parent):
        """ Creates the actual page content. """

        # Dummy implementation - override!
        panel = wx.Panel(parent, -1, style=wx.CLIP_CHILDREN)
        panel.SetBackgroundColour('yellow')

        return panel

#### EOF ######################################################################



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

  fs.tests.test_remote:  testcases for FS remote support utilities

"""

from fs.tests import FSTestCases, ThreadingTestCases

import unittest
import threading
import random
import time
import sys

from fs.remote import *

from fs import SEEK_END
from fs.wrapfs import WrapFS, wrap_fs_methods
from fs.tempfs import TempFS
from fs.path import *
from fs.local_functools import wraps

from six import PY3, b


class RemoteTempFS(TempFS):
    """
        Simple filesystem implementing setfilecontents
        for RemoteFileBuffer tests
    """
    def __repr__(self):
        return '<RemoteTempFS: %s>' % self._temp_dir

    def open(self, path, mode='rb', write_on_flush=True, **kwargs):
        if 'a' in mode or 'r' in mode or '+' in mode:
            f = super(RemoteTempFS, self).open(path, mode='rb', **kwargs)
            f = TellAfterCloseFile(f)
        else:
            f = None

        return RemoteFileBuffer(self,
                                path,
                                mode,
                                f,
                                write_on_flush=write_on_flush)

    def setcontents(self, path, data, encoding=None, errors=None, chunk_size=64*1024):
        f = super(RemoteTempFS, self).open(path, 'wb', encoding=encoding, errors=errors, chunk_size=chunk_size)
        if getattr(data, 'read', False):
            f.write(data.read())
        else:
            f.write(data)
        f.close()


class TellAfterCloseFile(object):
    """File-like object that allows calling tell() after it's been closed."""

    def __init__(self, file):
        self._finalpos = None
        self.file = file

    def close(self):
        if self._finalpos is None:
            self._finalpos = self.file.tell()
        self.file.close()

    def tell(self):
        if self._finalpos is not None:
            return self._finalpos
        return self.file.tell()

    def __getattr__(self, attr):
        return getattr(self.file, attr)


class TestRemoteFileBuffer(unittest.TestCase, FSTestCases, ThreadingTestCases):
    class FakeException(Exception): pass

    def setUp(self):
        self.fs = RemoteTempFS()
        self.original_setcontents = self.fs.setcontents

    def tearDown(self):
        self.fs.close()
        self.fakeOff()

    def fake_setcontents(self, path, content=b(''), chunk_size=16*1024):
        ''' Fake replacement for RemoteTempFS setcontents() '''
        raise self.FakeException("setcontents should not be called here!")

    def fakeOn(self):
        '''
            Turn on fake_setcontents(). When setcontents on RemoteTempFS
            is called, FakeException is raised and nothing is stored.
        '''
        self.fs.setcontents = self.fake_setcontents

    def fakeOff(self):
        ''' Switch off fake_setcontents(). '''
        self.fs.setcontents = self.original_setcontents

    def test_ondemand(self):
        '''
            Tests on-demand loading of remote content in RemoteFileBuffer
        '''
        contents = b("Tristatricettri stribrnych strikacek strikalo") + \
                   b("pres tristatricettri stribrnych strech.")
        f = self.fs.open('test.txt', 'wb')
        f.write(contents)
        f.close()

        # During following tests, no setcontents() should be called.
        self.fakeOn()

        f = self.fs.open('test.txt', 'rb')
        self.assertEquals(f.read(10), contents[:10])
        f.wrapped_file.seek(0, SEEK_END)
        self.assertEquals(f._rfile.tell(), 10)
        f.seek(20)
        self.assertEquals(f.tell(), 20)
        self.assertEquals(f._rfile.tell(), 20)
        f.seek(0, SEEK_END)
        self.assertEquals(f._rfile.tell(), len(contents))
        f.close()

        f = self.fs.open('test.txt', 'ab')
        self.assertEquals(f.tell(), len(contents))
        f.close()

        self.fakeOff()

        # Writing over the rfile edge
        f = self.fs.open('test.txt', 'wb+')
        self.assertEquals(f.tell(), 0)
        f.seek(len(contents) - 5)
        # Last 5 characters not loaded from remote file
        self.assertEquals(f._rfile.tell(), len(contents) - 5)
        # Confirm that last 5 characters are still in rfile buffer
        self.assertEquals(f._rfile.read(), contents[-5:])
        # Rollback position 5 characters before eof
        f._rfile.seek(len(contents[:-5]))
        # Write 10 new characters (will make contents longer for 5 chars)
        f.write(b('1234567890'))
        f.flush()
        # We are on the end of file (and buffer not serve anything anymore)
        self.assertEquals(f.read(), b(''))
        f.close()

        self.fakeOn()

        # Check if we wrote everything OK from
        # previous writing over the remote buffer edge
        f = self.fs.open('test.txt', 'rb')
        self.assertEquals(f.read(), contents[:-5] + b('1234567890'))
        f.close()

        self.fakeOff()

    def test_writeonflush(self):
        '''
            Test 'write_on_flush' switch of RemoteFileBuffer.
            When True, flush() should call setcontents and store
            to remote destination.
            When False, setcontents should be called only on close().
        '''
        self.fakeOn()
        f = self.fs.open('test.txt', 'wb', write_on_flush=True)
        f.write(b('Sample text'))
        self.assertRaises(self.FakeException, f.flush)
        f.write(b('Second sample text'))
        self.assertRaises(self.FakeException, f.close)
        self.fakeOff()
        f.close()
        self.fakeOn()

        f = self.fs.open('test.txt', 'wb', write_on_flush=False)
        f.write(b('Sample text'))
        # FakeException is not raised, because setcontents is not called
        f.flush()
        f.write(b('Second sample text'))
        self.assertRaises(self.FakeException, f.close)
        self.fakeOff()

    def test_flush_and_continue(self):
        '''
            This tests if partially loaded remote buffer can be flushed
            back to remote destination and opened file is still
            in good condition.
        '''
        contents = b("Zlutoucky kun upel dabelske ody.")
        contents2 = b('Ententyky dva spaliky cert vyletel z elektriky')

        f = self.fs.open('test.txt', 'wb')
        f.write(contents)
        f.close()

        f = self.fs.open('test.txt', 'rb+')
        # Check if we read just 10 characters
        self.assertEquals(f.read(10), contents[:10])
        self.assertEquals(f._rfile.tell(), 10)
        # Write garbage to file to mark it as _changed
        f.write(b('x'))
        # This should read the rest of file and store file back to again.
        f.flush()
        f.seek(0)
        # Try if we have unocrrupted file locally...
        self.assertEquals(f.read(), contents[:10] + b('x') + contents[11:])
        f.close()

        # And if we have uncorrupted file also on storage
        f = self.fs.open('test.txt', 'rb')
        self.assertEquals(f.read(), contents[:10] + b('x') + contents[11:])
        f.close()

        # Now try it again, but write garbage behind edge of remote file
        f = self.fs.open('test.txt', 'rb+')
        self.assertEquals(f.read(10), contents[:10])
        # Write garbage to file to mark it as _changed
        f.write(contents2)
        f.flush()
        f.seek(0)
        # Try if we have unocrrupted file locally...
        self.assertEquals(f.read(), contents[:10] + contents2)
        f.close()

        # And if we have uncorrupted file also on storage
        f = self.fs.open('test.txt', 'rb')
        self.assertEquals(f.read(), contents[:10] + contents2)
        f.close()


class TestCacheFS(unittest.TestCase,FSTestCases,ThreadingTestCases):
    """Test simple operation of CacheFS"""

    def setUp(self):
        self._check_interval = sys.getcheckinterval()
        sys.setcheckinterval(10)
        self.wrapped_fs = TempFS()
        self.fs = CacheFS(self.wrapped_fs,cache_timeout=0.01)

    def tearDown(self):
        self.fs.close()
        sys.setcheckinterval(self._check_interval)

    def test_values_are_used_from_cache(self):
        old_timeout = self.fs.cache_timeout
        self.fs.cache_timeout = None
        try:
            self.assertFalse(self.fs.isfile("hello"))
            self.wrapped_fs.setcontents("hello",b("world"))
            self.assertTrue(self.fs.isfile("hello"))
            self.wrapped_fs.remove("hello")
            self.assertTrue(self.fs.isfile("hello"))
            self.fs.clear_cache()
            self.assertFalse(self.fs.isfile("hello"))
        finally:
            self.fs.cache_timeout = old_timeout

    def test_values_are_updated_in_cache(self):
        old_timeout = self.fs.cache_timeout
        self.fs.cache_timeout = None
        try:
            self.assertFalse(self.fs.isfile("hello"))
            self.wrapped_fs.setcontents("hello",b("world"))
            self.assertTrue(self.fs.isfile("hello"))
            self.wrapped_fs.remove("hello")
            self.assertTrue(self.fs.isfile("hello"))
            self.wrapped_fs.setcontents("hello",b("world"))
            self.assertTrue(self.fs.isfile("hello"))
            self.fs.remove("hello")
            self.assertFalse(self.fs.isfile("hello"))
        finally:
            self.fs.cache_timeout = old_timeout



class TestConnectionManagerFS(unittest.TestCase,FSTestCases):#,ThreadingTestCases):
    """Test simple operation of ConnectionManagerFS"""

    def setUp(self):
        self._check_interval = sys.getcheckinterval()
        sys.setcheckinterval(10)
        self.fs = ConnectionManagerFS(TempFS())

    def tearDown(self):
        self.fs.close()
        sys.setcheckinterval(self._check_interval)


class DisconnectingFS(WrapFS):
    """FS subclass that raises lots of RemoteConnectionErrors."""

    def __init__(self,fs=None):
        if fs is None:
            fs = TempFS()
        self._connected = True
        self._continue = True
        self._bounce_thread = None
        super(DisconnectingFS,self).__init__(fs)
        if random.choice([True,False]):
            raise RemoteConnectionError("")
        self._bounce_thread = threading.Thread(target=self._bounce)
        self._bounce_thread.daemon = True
        self._bounce_thread.start()

    def __getstate__(self):
        state = super(DisconnectingFS,self).__getstate__()
        del state["_bounce_thread"]
        return state

    def __setstate__(self,state):
        super(DisconnectingFS,self).__setstate__(state)
        self._bounce_thread = threading.Thread(target=self._bounce)
        self._bounce_thread.daemon = True
        self._bounce_thread.start()

    def _bounce(self):
        while self._continue:
            time.sleep(random.random()*0.1)
            self._connected = not self._connected

    def setcontents(self, path, data=b(''), encoding=None, errors=None, chunk_size=64*1024):
        return self.wrapped_fs.setcontents(path, data, encoding=encoding, errors=errors, chunk_size=chunk_size)

    def close(self):
        if not self.closed:
            self._continue = False
            if self._bounce_thread is not None:
                self._bounce_thread.join()
            self._connected = True
            super(DisconnectingFS,self).close()


def disconnecting_wrapper(func):
    """Method wrapper to raise RemoteConnectionError if not connected."""
    @wraps(func)
    def wrapper(self,*args,**kwds):
        if not self._connected:
            raise RemoteConnectionError("")
        return func(self,*args,**kwds)
    return wrapper
DisconnectingFS = wrap_fs_methods(disconnecting_wrapper,DisconnectingFS,exclude=["close"])


class DisconnectRecoveryFS(WrapFS):
    """FS subclass that recovers from RemoteConnectionErrors by waiting."""
    pass
def recovery_wrapper(func):
    """Method wrapper to recover from RemoteConnectionErrors by waiting."""
    @wraps(func)
    def wrapper(self,*args,**kwds):
        while True:
            try:
                return func(self,*args,**kwds)
            except RemoteConnectionError:
                self.wrapped_fs.wait_for_connection()
    return wrapper
# this also checks that wrap_fs_methods works as a class decorator
DisconnectRecoveryFS = wrap_fs_methods(recovery_wrapper)(DisconnectRecoveryFS)


class TestConnectionManagerFS_disconnect(TestConnectionManagerFS):
    """Test ConnectionManagerFS's ability to wait for reconnection."""

    def setUp(self):
        self._check_interval = sys.getcheckinterval()
        sys.setcheckinterval(10)
        c_fs = ConnectionManagerFS(DisconnectingFS,poll_interval=0.1)
        self.fs = DisconnectRecoveryFS(c_fs)

    def tearDown(self):
        self.fs.close()
        sys.setcheckinterval(self._check_interval)

if __name__ == '__main__':
    unittest.main()



#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Float, Enum, Any, Property

from ..view import View

from ..item import Item

from ..editor_factory import EditorFactory

from ..basic_editor_factory import BasicEditorFactory

from .text_editor import TextEditor

from ..ui_traits import EditorStyle

from ..ui_editor import UIEditor

from ..toolkit import toolkit_object

#-------------------------------------------------------------------------
#  '_PopupEditor' class:
#-------------------------------------------------------------------------


class _PopupEditor(UIEditor):

    #-------------------------------------------------------------------------
    #  Creates the traits UI for the editor:
    #-------------------------------------------------------------------------

    def init_ui(self, parent):
        """ Creates the traits UI for the editor.
        """
        return self.object.edit_traits(view=self.base_view(),
                                       parent=parent)

    def base_view(self):
        """ Returns the View that allows the popup view to be displayed.
        """
        return View(
            Item(self.name,
                 show_label=False,
                 style='readonly',
                 editor=TextEditor(view=self.popup_view()),
                 padding=-4,
                 ),
            kind='subpanel'
        )

    def popup_view(self):
        """ Returns the popup View.
        """
        factory = self.factory
        item = Item(self.name,
                    show_label=False,
                    padding=-4,
                    style=factory.style,
                    height=factory.height,
                    width=factory.width)

        editor = factory.editor
        if editor is not None:
            if not isinstance(editor, EditorFactory):
                editor = editor()
            item.editor = editor

        return View(item, kind=factory.kind)

#-------------------------------------------------------------------------
#  'PopupEditor' class:
#-------------------------------------------------------------------------


class PopupEditor(BasicEditorFactory):

    # The class used to construct editor objects:
    klass = Property

    # The kind of popup to use:
    kind = Enum('popover', 'popup', 'info')

    # The editor to use for the pop-up view (can be None (use default editor),
    # an EditorFactory instance, or a callable that returns an EditorFactory
    # instance):
    editor = Any

    # The style of editor to use for the popup editor (same as Item.style):
    style = EditorStyle

    # The height of the popup (same as Item.height):
    height = Float(-1.0)

    # The width of the popup (same as Item.width):
    width = Float(-1.0)

    def _get_klass(self):
        """ The class used to construct editor objects.
        """
        return toolkit_object('popup_editor:_PopupEditor')



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought util package component>
#------------------------------------------------------------------------------
import wx

try:
    from scimath.units.unit_parser import unit_parser
except ImportError:
    unit_parser = None

from .default_renderer import DefaultRenderer

class UnitRenderer(DefaultRenderer):

    def DrawForeground(self, grid, attr, dc, rect, row, col, isSelected):
        dc.SetBackgroundMode(wx.TRANSPARENT)
        text = grid.model.GetValue(row, col)
        #print 'Rendering ', row, col, text, text.__class__
        dc.SetFont(self.font)
        dc.DrawText(text, rect.x+1, rect.y+1)

        return

    def DrawBackground(self, grid, attr, dc, rect, row, col, isSelected):
        """ Erases whatever is already in the cell by drawing over it.
        """
        # We have to set the clipping region on the grid's DC,
        # otherwise the text will spill over to the next cell
        dc.SetClippingRect(rect)

        # overwrite anything currently in the cell ...
        dc.SetBackgroundMode(wx.SOLID)

        dc.SetPen(wx.Pen(wx.WHITE, 1, wx.SOLID))

        text = grid.model.GetValue(row, col)

        if isSelected:
            dc.SetBrush(DefaultRenderer.selected_cells)
        elif unit_parser and unit_parser.parse_unit(text).is_valid():
            dc.SetBrush(DefaultRenderer.normal_cells)
        else:
            dc.SetBrush(DefaultRenderer.error_cells)

        dc.DrawRectangle(rect.x, rect.y, rect.width, rect.height)
        return
#-------------------------------------------------------------------------------

class MultiUnitRenderer(DefaultRenderer):

    def __init__( self, color="black", font="ARIAL", fontsize=8,
                  suppress_warnings=False):

        self.suppress_warnings = suppress_warnings
        DefaultRenderer.__init__( self, color, font, fontsize )

    def DrawForeground(self, grid, attr, dc, rect, row, col, isSelected):
        dc.SetBackgroundMode(wx.TRANSPARENT)
        text = grid.model.GetValue(row, col)
        dc.SetFont(self.font)
        dc.DrawText(text, rect.x+1, rect.y+1)

        return

    def DrawBackground(self, grid, attr, dc, rect, row, col, isSelected):
        """ Erases whatever is already in the cell by drawing over it.
        """
        # We have to set the clipping region on the grid's DC,
        # otherwise the text will spill over to the next cell
        dc.SetClippingRect(rect)

        # overwrite anything currently in the cell ...
        dc.SetBackgroundMode(wx.SOLID)

        dc.SetPen(wx.Pen(wx.WHITE, 1, wx.SOLID))

        text = grid.model.GetValue(row, col)
        if unit_parser:
            this_unit = unit_parser.parse_unit(text, self.suppress_warnings)
        else:
            this_unit = None

        # Todo - clean up this hardcoded logic/column position mess

        family = grid.model.GetValue(row, 6)

        # AI units of 'impedance ((kg/s)*(g/cc))' creates list of 3, not 2!
        try:
            family, other_text = family[:-1].split('(')
        except:
            family, other_text = family[:-1].split(' ')

        if unit_parser:
            other_unit = unit_parser.parse_unit(other_text, self.suppress_warnings)
            dimensionally_equivalent = this_unit.can_convert(other_unit)
        else:
            other_unit = None
            dimensionally_equivalent = False

        if isSelected:
            dc.SetBrush(DefaultRenderer.selected_cells)
        elif not this_unit or not this_unit.is_valid():
            dc.SetBrush(DefaultRenderer.error_cells)
        elif not dimensionally_equivalent:
            dc.SetBrush(DefaultRenderer.warn_cells)
        else:
            dc.SetBrush(DefaultRenderer.normal_cells)

        dc.DrawRectangle(rect.x, rect.y, rect.width, rect.height)
        return



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2006 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Dave Peterson <dpeterson@enthought.com>
#
#-----------------------------------------------------------------------------

""" Manages a singleton updater that acts as a global registry.

    Our goal is to enable the apptools.sweet_pickle framework to
    understand how pickled data should be treated during unpickling so
    that the resulting object hierarchy reflects the current versions
    of the object's classes -AND- that this should work no matter who
    is doing the unpickling.  This requires a global registry since
    there is no way to know in advance what objects are contained within
    a given pickle file, and thus no way to gather the required
    information to know how to treat the pickle data during unpickling.

    For example, a pickle of an Envisage project may contain many
    custom classes, instantiated at the behest of various plugins,
    which have gone through various versionings and refactorings.  But
    it is the project plugin that needs to unpickle these objects to
    'load' a project, not the other plugins that added those custom
    class instances into the project.


    This registry is used by the apptools.sweet_pickle framework's
    unpickler only by default.   That is, only if no updater was
    explicitly provided.


    It is important that users interact with the registry through the
    provided methods.  If they do not, then the reference they receive
    will be the one that was in place at the time of the import which
    may or MAY NOT be the current repository due to the way this
    framework manages the repository.
"""

try:
    import thread as _thread
except ImportError:
    import dummy_thread as _thread


##############################################################################
# function 'get_global_registry'
##############################################################################

def get_global_registry():
    """ Returns the global registry in a manner that allows for lazy
        instantiation.
    """
    global _global_registry, _global_registry_lock

    # Do we need to create the registry?
    if _global_registry is None:

        # We can only do this safely in a threaded situation by using a lock.
        # Note that the previous check for None doesn't guarantee we are
        # the only one trying to create an instance, so, we'll check for none
        # again once we acquire the lock and then only create the singleton
        # if there still isn't an instance.
        _global_registry_lock.acquire()
        try:
            if _global_registry is None:
                from updater import Updater
                _global_registry = Updater()
        finally:
            _global_registry_lock.release()

    return _global_registry


##############################################################################
# private function '_clear_global_registry'
##############################################################################

def _clear_global_registry():
    """ Clears out the current global registry.

        This exists purely to allow testing of the global registry and the
        apptools.sweet_pickle framework.  THIS METHOD SHOULD NEVER BE
        CALLED DURING NORMAL OPERATIONS!
    """
    global _global_registry
    _global_registry = None


##############################################################################
# private, but global, variables
##############################################################################

# The global singleton updater
_global_registry = None

# The lock used to make access to the global singleton thread safe
_global_registry_lock = _thread.allocate_lock()


#### EOF #####################################################################




#------------------------------------------------------------------------------
#
#  Copyright (c) 2006, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   01/05/2006
#
#------------------------------------------------------------------------------

""" Defines the tree-based Python value editor and the value editor factory.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Instance, Int, false

from .tree_editor import TreeEditor
from ..view import View
from ..item import Item

from ..value_tree import RootNode, value_tree_nodes

from ..editor_factory import EditorFactory

from ..editor import Editor

#-------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------


class _ValueEditor(Editor):
    """ Simple style of editor for values, which displays a tree.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Is the editor read only?
    readonly = false

    # The root node of the value tree
    root = Instance(RootNode)

    # Is the value editor scrollable? This values overrides the default.
    scrollable = True

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.update_editor()
        editor = TreeEditor(
            auto_open=self.factory.auto_open,
            hide_root=True,
            editable=False,
            nodes=value_tree_nodes
        )
        self._ui = self.edit_traits(parent=parent, view=View(
            Item('root',
                 show_label=False,
                 editor=editor
                 ),
            kind='subpanel'
        )
        )
        self._ui.parent = self.ui
        self.control = self._ui.control

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes external to the
            editor.
        """
        self.root = RootNode(name='',
                             value=self.value,
                             readonly=self.readonly)

    #-------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        self._ui.dispose()

        super(_ValueEditor, self).dispose()

    #-------------------------------------------------------------------------
    #  Returns the editor's control for indicating error status:
    #-------------------------------------------------------------------------

    def get_error_control(self):
        """ Returns the editor's control for indicating error status.
        """
        return self._ui.get_error_controls()

#-------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------


class ToolkitEditorFactory(EditorFactory):
    """ Editor factory for tree-based value editors.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Number of tree levels to automatically open
    auto_open = Int(2)

# Define the ValueEditor class.
ValueEditor = ToolkitEditorFactory

#--EOF-------------------------------------------------------------------------



#!/usr/bin/env python
import cStringIO

import numpy as np
from PIL import Image


def get_numpy_from_data(data):
    image = Image.open(cStringIO.StringIO(data))
    if image.mode != "RGBA":
        image = image.convert("RGBA")
    arr = np.array(image)
    return arr


def get_rect(w, h, color=None, opaque_border=False):
    arr = np.empty((h, w, 4), np.uint8)

    # just some indexes to keep track of which byte is which
    R, G, B, A = range(4)

    if color is None:
        red, green, blue, alpha = (35, 142,  35, 128)
    else:
        red, green, blue, alpha = color
    # initialize all pixel values to the values passed in
    arr[:,:,R] = red
    arr[:,:,G] = green
    arr[:,:,B] = blue
    arr[:,:,A] = alpha

    if opaque_border:
        # Set the alpha for the border pixels to be fully opaque
        arr[0,   0:w, A] = 255  # first row
        arr[h-1, 0:w, A] = 255  # last row
        arr[0:h, 0,   A] = 255  # first col
        arr[0:h, w-1, A] = 255  # last col

    return arr


def get_checkerboard(w, h, sq=16):
    # Algorithm from http://stackoverflow.com/questions/2169478/how-to-make-a-checkerboard-in-numpy
    color1 = (64, 64, 64, 32)
    color2 = (255, 255, 255, 32)
    coords = np.ogrid[0:h, 0:w]
    idx = (coords[0] // sq + coords[1] // sq) & 1
    vals = np.array([color1, color2], dtype=np.uint8)
    arr = vals[idx]
    return arr


def get_square(size):
    return get_rect(size, size)


def get_image_from_data(array):
    image = Image.fromarray(array)
    return image


def save_image(array, filename):
    image = get_image_from_data(array)
    image.save(filename)


if __name__ == "__main__":
    # Create a Jumpman peanut allergy mask
    hx = 0
    hy = 6
    xb = lambda x:x+0x30+hx
    yb = lambda y:2*y+0x20+hy

    def is_allergic(x, y, hx, hy):
        return (x + 0x30 + hx) & 0x1f < 7 or (2 * y + 0x20 + hy) & 0x1f < 5
    w = 160
    h = 88
    good = (64, 178, 200, 100)
    bad = (203, 144, 161, 100)
    playfield = get_rect(w, h, good)
    for x in range(w):
        for y in range(h):
            if is_allergic(x, y, hx, hy):
                playfield[y, x] = bad
    save_image(playfield, "bad_peanut.png")

    print "X bomb location,hx=0,hx=16,hx=f0,hx=fc"
    hy = 6
    for x in range(w):
        #print "$%02x,%s" % (x, "OK" if not is_allergic(x, 0, hx, 6) else "BAD!")
        badlist = []
        for hx in [0, 16, -16, -4]:
            badlist.append("ok" if not is_allergic(x, 0, hx, hy) else "BAD!")
        print "$%02x,%s" % (x, ",".join(badlist))

    print "Y bomb location,hy=0,hy=2,hy=4,hy=6,hy=f6"
    hx = 0
    for y in range(h):
        #print "$%02x,%s" % (x, "OK" if not is_allergic(x, 0, hx, 6) else "BAD!")
        badlist = []
        for hy in [0, 2, 4, 6, -10]:
            badlist.append("ok" if not is_allergic(0, y, hx, hy) else "BAD!")
        print "$%02x,%s" % (y, ",".join(badlist))



import unittest

from ..type_registry import TypeRegistry

from .dummies import A, B, C, D, Mixed, Abstract, Concrete, ConcreteSubclass


class TestTypeRegistry(unittest.TestCase):

    def setUp(self):
        self.registry = TypeRegistry()

    def test_deferred_push(self):
        self.registry.push('dummies:A', 'A')
        self.registry.push('dummies:C', 'C')
        self.assertEqual(self.registry.lookup_by_type(A), 'A')
        self.assertEqual(self.registry.lookup_by_type(B), 'A')
        self.assertEqual(self.registry.lookup_by_type(C), 'C')
        self.assertRaises(KeyError, self.registry.lookup_by_type, D)

    def test_greedy_push(self):
        self.registry.push(A, 'A')
        self.registry.push(C, 'C')
        self.assertEqual(self.registry.lookup_by_type(A), 'A')
        self.assertEqual(self.registry.lookup_by_type(B), 'A')
        self.assertEqual(self.registry.lookup_by_type(C), 'C')
        self.assertRaises(KeyError, self.registry.lookup_by_type, D)

    def test_pop_by_name_from_name(self):
        self.registry.push('dummies:A', 'A')
        self.registry.pop('dummies:A')
        self.assertRaises(KeyError, self.registry.lookup_by_type, A)

    def test_pop_by_name_from_type(self):
        self.registry.push(A, 'A')
        self.registry.pop('dummies:A')
        self.assertRaises(KeyError, self.registry.lookup_by_type, A)

    def test_pop_by_type_from_name(self):
        self.registry.push('dummies:A', 'A')
        self.registry.pop(A)
        self.assertRaises(KeyError, self.registry.lookup_by_type, A)

    def test_pop_by_type_from_type(self):
        self.registry.push(A, 'A')
        self.registry.pop(A)
        self.assertRaises(KeyError, self.registry.lookup_by_type, A)

    def test_mro(self):
        self.registry.push('dummies:A', 'A')
        self.registry.push('dummies:D', 'D')
        self.assertEqual(self.registry.lookup_by_type(Mixed), 'A')

    def test_lookup_instance(self):
        self.registry.push(A, 'A')
        self.registry.push(C, 'C')
        self.assertEqual(self.registry.lookup(A()), 'A')
        self.assertEqual(self.registry.lookup(B()), 'A')
        self.assertEqual(self.registry.lookup(C()), 'C')
        self.assertRaises(KeyError, self.registry.lookup, D())

    def test_lookup_all(self):
        self.registry.push(A, 'A')
        self.registry.push(C, 'C')
        self.assertEqual(self.registry.lookup_all(A()), ['A'])
        self.assertEqual(self.registry.lookup_all(B()), ['A'])
        self.registry.push(A, 'A2')
        self.assertEqual(self.registry.lookup_all(A()), ['A', 'A2'])
        self.assertEqual(self.registry.lookup_all(B()), ['A', 'A2'])

    def test_abc(self):
        self.registry.push_abc(Abstract, 'Abstract')
        self.assertEqual(self.registry.lookup_by_type(Concrete), 'Abstract')
        self.assertEqual(self.registry.lookup_by_type(ConcreteSubclass),
            'Abstract')

    def test_stack_type(self):
        self.registry.push(A, 'A1')
        self.registry.push(A, 'A2')
        self.assertEqual(self.registry.lookup_by_type(A), 'A2')
        self.registry.pop(A)
        self.assertEqual(self.registry.lookup_by_type(A), 'A1')
        self.registry.pop(A)
        self.assertRaises(KeyError, self.registry.lookup_by_type, A)
        self.assertRaises(KeyError, self.registry.pop, A)
        self.assertRaises(KeyError, self.registry.pop, 'dummies:A')
        self.assertEqual(self.registry.type_map, {})
        self.assertEqual(self.registry.name_map, {})
        self.assertEqual(self.registry.abc_map, {})

    def test_stack_mixed_type_and_name(self):
        self.registry.push(A, 'A1')
        self.registry.push('dummies:A', 'A2')
        self.assertEqual(self.registry.lookup_by_type(A), 'A2')
        self.registry.pop(A)
        self.assertEqual(self.registry.lookup_by_type(A), 'A1')
        self.registry.pop('dummies:A')
        self.assertRaises(KeyError, self.registry.lookup_by_type, A)
        self.assertRaises(KeyError, self.registry.pop, A)
        self.assertRaises(KeyError, self.registry.pop, 'dummies:A')
        self.assertEqual(self.registry.type_map, {})
        self.assertEqual(self.registry.name_map, {})
        self.assertEqual(self.registry.abc_map, {})

    def test_stack_abc(self):
        self.registry.push_abc(Abstract, 'Abstract1')
        self.registry.push_abc(Abstract, 'Abstract2')
        self.assertEqual(self.registry.lookup_by_type(Concrete), 'Abstract2')
        self.registry.pop(Abstract)
        self.assertEqual(self.registry.lookup_by_type(Concrete), 'Abstract1')
        self.registry.pop(Abstract)
        self.assertRaises(KeyError, self.registry.lookup_by_type, Concrete)
        self.assertRaises(KeyError, self.registry.pop, Abstract)
        self.assertEqual(self.registry.type_map, {})
        self.assertEqual(self.registry.name_map, {})
        self.assertEqual(self.registry.abc_map, {})



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   06/13/2007
#
#-------------------------------------------------------------------------

""" Traits UI simple, read-only single line text editor with a themed
    (i.e. image) background.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from traits.api \
    import Enum, Instance, Bool, Dict, Str, Any, Property, TraitError, \
    cached_property

from traitsui.ui_traits \
    import ATheme

from traitsui.wx.editor \
    import Editor

from traitsui.wx.editor_factory \
    import EditorFactory

from pyface.image_resource \
    import ImageResource

from image_slice import paint_parent, ImageSlice

from constants \
    import OKColor, ErrorColor

from helper \
    import disconnect_no_id, BufferDC

from themed_control \
    import ThemedControl


# Define a reusable, default ImageSlice object:
default_image_slice = ImageSlice()


#-------------------------------------------------------------------------
#  Define a simple identity mapping:
#-------------------------------------------------------------------------

class _Identity(object):
    """ A simple indentity mapping.
    """

    def __call__(self, value):
        return value

#-------------------------------------------------------------------------
#  'ThemedTextEditor' class:
#-------------------------------------------------------------------------


class ThemedTextEditor(EditorFactory):
    """ Traits UI simple, single line text editor with a themed (i.e. image)
        background.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The background theme image to display:
    theme = ATheme

    # Dictionary that maps user input to other values:
    mapping = Dict(Str, Any)

    # Is user input set on every keystroke?
    auto_set = Bool(True)

    # Is user input set when the Enter key is pressed?
    enter_set = Bool(False)

    # Is user input unreadable? (e.g., for a password)
    password = Bool(False)

    # Function to evaluate textual user input:
    evaluate = Any

    # The object trait containing the function used to evaluate user input:
    evaluate_name = Str

    #-------------------------------------------------------------------------
    #  'Editor' factory methods:
    #-------------------------------------------------------------------------

    def simple_editor(self, ui, object, name, description, parent):
        return _ThemedTextEditor(parent,
                                 factory=self,
                                 ui=ui,
                                 object=object,
                                 name=name,
                                 description=description)

    def custom_editor(self, ui, object, name, description, parent):
        return _ThemedTextEditor(parent,
                                 factory=self,
                                 ui=ui,
                                 object=object,
                                 name=name,
                                 description=description)

    def text_editor(self, ui, object, name, description, parent):
        return _ThemedTextEditor(parent,
                                 factory=self,
                                 ui=ui,
                                 object=object,
                                 name=name,
                                 description=description)

    def readonly_editor(self, ui, object, name, description, parent):
        return _ReadonlyTextEditor(parent,
                                   factory=self,
                                   ui=ui,
                                   object=object,
                                   name=name,
                                   description=description)

#-------------------------------------------------------------------------
#  '_ThemedTextEditor' class:
#-------------------------------------------------------------------------


class _ThemedTextEditor(Editor):
    """ Traits UI simple, single line text editor with a themed (i.e. image
        background).
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Function used to evaluate textual user input:
    evaluate = Any

    # The text alignment to use:
    alignment = Property

    # The image slice to use:
    image_slice = Property

    #-- Class Variables ------------------------------------------------------

    text_styles = {
        'default': wx.TE_LEFT,
        'left': wx.TE_LEFT,
        'center': wx.TE_CENTRE,
        'right': wx.TE_RIGHT
    }

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        factory = self.factory
        evaluate = factory.evaluate
        if evaluate is None:
            handler = self.object.trait(self.name).handler
            evaluate = getattr(handler, 'evaluate', None)
            if evaluate is None:
                evaluate = _Identity()
        self.evaluate = evaluate
        self.sync_value(factory.evaluate_name, 'evaluate', 'from')

        padding_x = padding_y = 0
        if factory.theme is not None:
            slice = self.image_slice
            padding_x = slice.xleft + slice.xright
            padding_y = slice.xtop + slice.xbottom

        self.control = control = wx.Window(parent, -1,
                                           size=wx.Size(padding_x + 70, padding_y + 20),
                                           style=wx.FULL_REPAINT_ON_RESIZE | wx.WANTS_CHARS)

        self._text_size = None

        # Set up the painting event handlers:
        wx.EVT_ERASE_BACKGROUND(control, self._erase_background)
        wx.EVT_PAINT(control, self._on_paint)
        wx.EVT_CHAR(control, self._inactive_key_entered)

        # Handle 'focus' events:
        wx.EVT_SET_FOCUS(control, self._set_focus)
        wx.EVT_LEFT_UP(control, self._set_focus)

        # Handle 'resize' events:
        wx.EVT_SIZE(control, self._resize)

        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        # Remove all of the wx event listeners:
        disconnect_no_id(
            self.control,
            wx.EVT_ERASE_BACKGROUND,
            wx.EVT_PAINT,
            wx.EVT_CHAR,
            wx.EVT_SET_FOCUS,
            wx.EVT_LEFT_UP,
            wx.EVT_SIZE)

        super(_ThemedTextEditor, self).dispose()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        if self._text is None:
            self._refresh()
            return

        if self._get_user_value() != self.value:
            self._no_update = True
            self._text.SetValue(self.str_value)
            self._no_update = False

        if self._error is not None:
            self._error = None
            self.ui.errors -= 1
            self._text.SetBackgroundColour(self.ok_color)
            self._text.Refresh()

    #-------------------------------------------------------------------------
    #  Handles the user entering input data in the edit control:
    #-------------------------------------------------------------------------

    def update_object(self, event):
        """ Handles the user entering input data in the edit control.
        """
        if not self._no_update:
            try:
                self.value = self._get_user_value()
                self._text.SetBackgroundColour(OKColor)
                self._text.Refresh()

                if self._error is not None:
                    self._error = None
                    self.ui.errors -= 1
                return True

            except TraitError as excp:
                return False

    #-------------------------------------------------------------------------
    #  Gets the actual value corresponding to what the user typed:
    #-------------------------------------------------------------------------

    def _get_user_value(self):
        """ Gets the actual value corresponding to what the user typed.
        """
        value = self._text.GetValue()
        try:
            value = self.evaluate(value)
        except:
            pass

        return self.factory.mapping.get(value, value)

    #-------------------------------------------------------------------------
    #  Handles an error that occurs while setting the object's trait value:
    #-------------------------------------------------------------------------

    def error(self, excp):
        """ Handles an error that occurs while setting the object's trait value.
        """
        self._text.SetBackgroundColour(ErrorColor)
        self._text.Refresh()
        wx.Bell()

        if self._error is None:
            self._error = True
            self.ui.errors += 1

    #-- Private Methods ------------------------------------------------------

    def _pop_up_text(self):
        """ Pop-up a text control to allow the user to enter a value using
            the keyboard.
        """
        control = self.control
        factory = self.factory
        style = (self.text_styles[self.alignment] | wx.TE_PROCESS_ENTER)
        if factory.password:
            style |= wx.TE_PASSWORD

        self._text = text = wx.TextCtrl(control, -1, self.str_value,
                                        style=style)
        slice = self.image_slice
        wdx, wdy = control.GetClientSize()
        tdx, tdy = text.GetSize()
        text.SetPosition(wx.Point(
            slice.xleft, ((wdy + slice.xtop - slice.xbottom - tdy) / 2) + 1))
        text.SetSize(wx.Size(wdx - slice.xleft - slice.xright, tdy))
        text.SetSelection(-1, -1)
        text.SetFocus()

        wx.EVT_KILL_FOCUS(text, self._text_completed)
        wx.EVT_CHAR(text, self._key_entered)
        wx.EVT_TEXT_ENTER(control, text.GetId(), self.update_object)

        if factory.auto_set and (not factory.is_grid_cell):
            wx.EVT_TEXT(control, text.GetId(), self.update_object)

    def _destroy_text(self):
        """ Destroys the current text control.
        """
        self.control.DestroyChildren()
        self._text = None

    def _refresh(self):
        """ Refreshes the contents of the control.
        """
        if self._text_size is not None:
            self.control.RefreshRect(wx.Rect(*self._get_text_bounds()),
                                     False)
            self._text_size = None

        self.control.RefreshRect(wx.Rect(*self._get_text_bounds()), False)

    def _get_text_size(self):
        """ Returns the text size information for the window.
        """
        if self._text_size is None:
            self._text_size = self.control.GetFullTextExtent(
                self._get_text() or 'M')

        return self._text_size

    def _get_text_bounds(self):
        """ Get the window bounds of where the current text should be
            displayed.
        """
        tdx, tdy, descent, leading = self._get_text_size()
        wdx, wdy = self.control.GetClientSizeTuple()
        slice = self.image_slice
        ady = wdy - slice.xtop - slice.xbottom
        ty = slice.xtop + ((ady - (tdy - descent)) / 2) - 1
        alignment = self.alignment
        if alignment == 'center':
            adx = wdx - slice.xleft - slice.xright
            tx = slice.xleft + (adx - tdx) / 2
        elif alignment == 'right':
            tx = wdx - tdx - slice.xright - 4
        else:
            tx = slice.xleft + 4

        return (tx, ty, tdx, tdy)

    def _get_text(self):
        """ Returns the current text to display.
        """
        if self.factory.password:
            return '*' * len(self.str_value)

        return self.str_value

    #-- Property Implementations ---------------------------------------------

    @cached_property
    def _get_alignment(self):
        theme = self.factory.theme
        if theme is not None:
            return theme.alignment

        return 'left'

    @cached_property
    def _get_image_slice(self):
        theme = self.factory.theme
        if theme is not None:
            return theme.image_slice or default_image_slice

        return default_image_slice

    #-- wxPython Event Handlers ----------------------------------------------

    def _erase_background(self, event):
        """ Do not erase the background here (do it in the 'on_paint' handler).
        """
        pass

    def _on_paint(self, event):
        """ Paint the background using the associated ImageSlice object.
        """
        control = self.control
        dc = BufferDC(control)
        slice = paint_parent(dc, control)
        slice2 = self.image_slice
        if slice2 is not default_image_slice:
            wdx, wdy = control.GetClientSizeTuple()
            slice2.fill(dc, 0, 0, wdx, wdy, True)
            slice = slice2
        elif slice is None:
            slice = slice2
        dc.SetBackgroundMode(wx.TRANSPARENT)
        dc.SetTextForeground(slice.content_color)
        dc.SetFont(control.GetFont())
        tx, ty, tdx, tdy = self._get_text_bounds()
        dc.DrawText(self._get_text(), tx, ty)
        dc.copy()

    def _resize(self, event):
        """ Handles the control being resized.
        """
        if self._text is not None:
            self._text.SetSize(self.control.GetSize())

    def _set_focus(self, event):
        """ Handle the control getting the keyboard focus.
        """
        if self._text is None:
            self._pop_up_text()

        event.Skip()

    def _text_completed(self, event):
        """ Handles the user transferring focus out of the text control.
        """
        if isinstance(event, wx.FocusEvent):
            event.Skip()
        if self.update_object(event):
            self._destroy_text()

    def _key_entered(self, event):
        """ Handles individual key strokes while the text control is active.
        """
        key_code = event.GetKeyCode()
        if key_code == wx.WXK_ESCAPE:
            self._destroy_text()
            return

        if key_code == wx.WXK_TAB:
            if self.update_object(event):
                if event.ShiftDown():
                    self.control.Navigate(0)
                else:
                    self.control.Navigate()

            return

        event.Skip()

    def _inactive_key_entered(self, event):
        """ Handles individual key strokes while the text control is inactive.
        """
        if event.GetKeyCode() == wx.WXK_RETURN:
            if self._text is None:
                self._pop_up_text()

            return

        event.Skip()

#-------------------------------------------------------------------------
#  '_ReadonlyTextEditor' class:
#-------------------------------------------------------------------------


class _ReadonlyTextEditor(Editor):
    """ Traits UI simple, read-only single line text view with a themed (i.e.
        image background).
    """

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self._control = ThemedControl(theme=self.factory.theme)
        self.control = self._control.create_control(parent)
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        self._control.text = self.value

        # Make sure the control is sized correctly:
        self.control.SetMinSize(self._control.best_size)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Enthought, Inc.
#
#------------------------------------------------------------------------------

""" A status bar manager realizes itself in a status bar control.
"""

# Major package imports.
import wx
from wx.lib.stattext import GenStaticText  # standard static text can't set background color on some platforms

# Enthought library imports.
from traits.api import Any, Unicode, Int, Float, Bool, Event
from pyface.action.api import StatusBarManager


class FrameworkStatusBar(wx.StatusBar):
    """A custom status bar for displaying the application version in the bottom
    right corner."""

    def __init__(self, parent, widths, error_delay):
        wx.StatusBar.__init__(self, parent, -1)
        self.error_delay = error_delay

        self.SetFieldsCount(len(widths))
        self.SetStatusWidths(widths)

        self.field_widths = list(widths)
        self.error_control = GenStaticText(self, -1)
        self.error_control.Hide()
        self.message_expire_time = 0
        self.reposition_controls()

        self.expire_timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.on_timer, self.expire_timer)
        self.Bind(wx.EVT_SIZE, self.on_size)

        f = self.GetFont()
        self.error_control.SetFont(wx.Font(f.GetPointSize(), f.GetFamily(),
                                 f.GetStyle(), wx.BOLD, f.GetUnderlined(),
                                 f.GetFaceName(), f.GetEncoding()))
        self.error_control.SetForegroundColour((255, 255, 255))
        self.error_control.SetBackgroundColour((211, 72, 54))

    def on_size(self, event):
        self.reposition_controls()

    def on_timer(self, event):
        self.error_control.Hide()

    def reposition_controls(self):
        rect = self.GetFieldRect(0)
        self.error_control.SetRect(rect)

    def show_error(self, text):
        self.error_control.SetLabel(" " + text)
        rect = self.GetFieldRect(0)
        self.error_control.SetRect(rect)
        self.error_control.Show()
        self.expire_timer.Start(self.error_delay * 1000, oneShot=True)


class FrameworkStatusBarManager(StatusBarManager):
    """ A status bar manager realizes itself in a status bar control. """

    # The message displayed in the first field of the status bar.
    message = Unicode

    message_width = Int(600)

    # The message to be displayed in debug field of the status bar
    command_help = Unicode

    command_help_width = Int(-1)

    # The message to be displayed in debug field of the status bar
    debug = Unicode

    debug_width = Int(200)

    # An error message to be displayed in the first field, and also for a
    # minimum time
    error = Event

    # Number of seconds to display an error message before the status bar is
    # freed up for normal messages
    error_delay = Int(3)

    # The toolkit-specific control that represents the status bar.
    status_bar = Any

    ###########################################################################
    # 'StatusBarManager' interface.
    ###########################################################################

    def create_status_bar(self, parent):
        """ Creates a status bar. """

        if self.status_bar is None:
            self.status_bar = FrameworkStatusBar(parent, [self.message_width, self.command_help_width, self.debug_width], self.error_delay)
            self.status_bar._pyface_control = self
            self.status_bar.SetStatusText(self.message, 0)
            self.status_bar.SetStatusText(self.command_help, 1)
            self.status_bar.SetStatusText(self.debug, 2)

        return self.status_bar

    def remove_status_bar(self, parent):
        """ Creates a status bar. """

        if self.status_bar is not None:
            self.status_bar.Destroy()
            self.status_bar._pyface_control = None
            self.status_bar = None

    ###########################################################################
    # Trait event handlers.
    ###########################################################################

    def _message_changed(self):
        """ Sets the text displayed on the status bar. """

        if self.status_bar is not None:
            self.status_bar.SetStatusText(self.message, 0)

        return

    def _command_help_changed(self):
        """ Sets the text displayed on the status bar. """

        if self.status_bar is not None:
            self.status_bar.SetStatusText(self.command_help, 1)

        return

    def _debug_changed(self):
        """ Sets the text displayed on the status bar. """

        if self.status_bar is not None:
            self.status_bar.SetStatusText(self.debug, 2)

        return

    def _error_fired(self, message):
        """Displays an error message and doesn't allow the status bar to be
        updated until a minimum time passes.
        """
        self.status_bar.show_error(message)



#!/usr/bin/env python

"""
Code for workign with particle fiels in mkz

Only handles reading for now
"""
# for py2/3 compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import zipfile
import base64
from datetime import datetime

import numpy as np

from post_gnome import nc_particles

file_attributes = nc_particles.file_attributes
var_attributes = nc_particles.var_attributes


class Writer(object):
    """
    class to write kmz files sutable for Google Earth
    """
    time_formatter = '%m/%d/%Y %H:%M'

    def __init__(self,
                 filename,
                 num_timesteps=None,
                 ref_time=None,
                 file_attributes=file_attributes,
                 var_attributes=var_attributes
                 ):

        """
        create a kmz_particle file Writer

        Creates the kml / kmz file, opens it for writing,
        writes the global attributes, creates required variables etc.

        :param filename: name of kmz file to open - if it exists,
                         it will be written over!

        :param num_timesteps=None: number of timesteps that will be output. Must be defined for netcdf3.
                                   Can be None for netcdf4
        :type num_timesteps: integer

        :param ref_time=None: reference time for time units (i.e. seconds since..).
                              If None, the first time used will be used.
        :type ref_time: datetime object

        :param file_attributes: keys and values for teh file-level attributes.
                                Defaults to the set defined in this module.
        :type file_attributes: dict

        :param var_attributes: dist of variable names, and the keys and values for variable
                               attributes.
                               Defaults to the set defined in this module.
        :type var_attributes: dict

        :param nc_version=3: version of netcdf to use -- must be 3 or 4. If 4, some extra
                             features are enabled.
        :type nc_version: integer
        """

        # strip off the .kml or .kmz
        filename = filename.rstrip(".kml").rstrip(".kmz")

        self.filename = filename + ".kmz"
        self.kml_name = os.path.split(filename)[-1] + ".kml"

        self.num_timesteps = num_timesteps
        self.ref_time = ref_time

        self.file_attributes = file_attributes
        self.var_attributes = var_attributes

        # create a list to hold what will be the contents of the kml
        self.kml = [header_template.format(caveat=caveat,
                                           kml_name=self.kml_name,
                                           # fixme: this is only doing now!
                                           valid_timestring=datetime.now().strftime(self.time_formatter),
                                           issued_timestring=datetime.now().strftime(self.time_formatter),
                                           )]
        # # fixme: put real data in here from nc_file?
        # self.kml = [header_template.format(caveat=caveat,
        #                                    kml_name=self.kml_name,
        #                                    valid_timestring=model_start_time.strftime(self.time_formatter),
        #                                    issued_timestring=datetime.now().strftime(self.time_formatter),
        #                                    )]

        # # Global attributes
        # # put some of this in the kml file?
        # for (name, value) in self.file_attributes.items():
        #     setattr(nc, name, value)
        self.closed = False





    def write_timestep(self, timestamp, timestep, data, uncertain=False):
        """
        write the data for a timestep

        :param timestamp: the time stamp of the timestep
        :type timestamp: datetime object

        :param timestep: the timestep between this and the next data point
        :type timestep: datetime object

        :param data: dict of data arrays -- all parameters for a single time step
        :type data: dict

        :param uncertain=False: Is this an uncertaintly run?
        :type uncertain: bool
        """

        start_time = timestamp.isoformat()
        end_time = (timestamp + timestep).isoformat()


        positions = np.c_[data['longitude'], data['latitude']]

        try:
            in_water = 2
            on_land = 3

            water_positions = positions[data['status_codes'] == in_water]
            beached_positions = positions[data['status_codes'] == on_land]
        except KeyError:
            water_positions = positions
            beached_positions = np.zeros((0, 2), dtype=np.float64)

        self.kml.append(build_one_timestep(water_positions,
                                           beached_positions,
                                           start_time,
                                           end_time,
                                           uncertain
                                           ))

    def close(self):
        """
        close the kmz file

        This forces the write of the file
        """
        if not self.closed:
            self.kml.append(footer)
            with zipfile.ZipFile(self.filename, 'w', compression=zipfile.ZIP_DEFLATED) as kmzfile:
                kmzfile.writestr('dot.png', base64.b64decode(DOT))
                kmzfile.writestr('x.png', base64.b64decode(X))
                # write the kml file
                kmzfile.writestr(self.kml_name, "".join(self.kml).encode('utf8'))
            self.closed = True
            return True
        else:
            return False

    def __del__(self):
        """ make sure to close the netcdf file """
        # anything to close?
        self.close()



class Reader(object):
    """
    Class to handle reading a nc_particle file

    (such as those written by GNOME or the Writer class above)
    """
    def __init__(self, kml_file):
        """
        initialize a file reader.

        :param kml_file: the kml/kmz file to read.
        :type kml_file: string

        """
        raise NotImplementedError

    @property
    def variables(self):
        """
        return the names of all the variables associated with the particles
        """
        raise NotImplementedError

    def __str__(self):
        return ("kml_particles Reader object:\n"
                "variables: {}\n"
                "number of timesteps: {}\n"
                ).format(self.variables, len(self.times))

    def get_all_timesteps(self, variables=['latitude', 'longitude']):
        """
        returns the requested variables data from all timesteps as a
        dictionary keyed by the variable names

        :param variables: the variables desired as a list string names.
                          Defaults to ['latitude','longitude']
        :type variables: list of strings

        :returns data: returns a dict of arrays -- the keys are the
                       variable names, and the values are numpy arrays
                        of the data. The arrays are the flattened ragged
                        array of data.
        """
        raise NotImplementedError

    def get_units(self, variable):
        """
        return the units of the given variable

        :param variable: name of the variable for which the units are required
        :type variable: string
        """
        raise NotImplementedError

    def get_attributes(self, variable):
        """
        return all the attributes of the given variable

        :param variable: name of the variable for which the attributes are required
        :type variable: string
        """
        raise NotImplementedError

    def get_timestep(self, timestep, variables=['latitude', 'longitude']):
        """
        returns the requested variables data from a given timestep as a
        dictionary keyed by the variable names

        :param variables: The variables desired as a list string names.
                          Defaults to ['latitude','longitude']
        :type variables: list of strings

        :returns data: returns a dict of arrays -- the keys are the
                       variable names, and the values are numpy arrays
                       of the data.
        """
        raise NotImplementedError

    def get_individual_trajectory(self, particle_id, variables=['latitude', 'longitude']):
        """
        returns the requested variables from trajectory of an individual particle

        note: this is inefficient -- it has to read the entire file to get it.
        """
        raise NotImplementedError

    def close(self):
        """
        close the kml file

        -- anything to be done?
        """
        pass


    def __del__(self):
        """ make sure to close the file """
        self.close()


def nc2kmz(nc_file, kmz_file=None):
    """
    convert a nc_particles file to kmz

    :param nc_file: name of nertcdf file to read

    :param kmz_file=None: name of kmz file to write. If None, the nc_file's name wil be used, with .kmz as teh extansion.

    """

    if kmz_file is None:
        root = nc_file
        root = root[:-3] if root.endswith(".nc") else root
        kmz_file = root + ".kmz"

    reader = nc_particles.Reader(nc_file)

    # create a kmz writer:
    writer = Writer(kmz_file)
    variables = reader.variables
    # loop to read / write the data
    for step, time in enumerate(reader.times):
        try:
            timestep = reader.times[step + 1] - time
        except IndexError:
            timestep = reader.times[-1] - reader.times[-2]
        # get the data
        data = reader.get_timestep(step, variables)
        writer.write_timestep(time, timestep, data, uncertain=False)
    writer.close()

    return kmz_file


# Templates for the kmz files

caveat = ("This trajectory was produced by GNOME (General NOAA Operational Modeling",
          " Environment), and should be used for educational and planning purposes only",
          "--not for a real response. In the event of an oil or chemical spill in U.S.",
          "waters, contact the U.S. Coast Guard National Response Center at 1-800-424-8802."
          )

# The kml templates:
header_template = """<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
  <Document>
    <name>{kml_name}</name>
    <open>1</open>
    <description><![CDATA[<b>Valid for:</b> {valid_timestring}<br>
                          <b>Issued:</b>{issued_timestring} <br>
                          {caveat}]]>
    </description>

    <Style id="RedDotIcon">
      <IconStyle>
         <scale>0.2</scale>
         <color>ff0000ff</color>
         <Icon>
            <href>dot.png</href>
         </Icon>
          <hotSpot x="0.5"  y="0.5" xunits="fraction" yunits="fraction"/>
      </IconStyle>
      <LabelStyle>
         <color>00000000</color>
      </LabelStyle>
    </Style>

    <Style id="BlackDotIcon">
      <IconStyle>
         <scale>0.2</scale>
         <Icon>
            <href>dot.png</href>
         </Icon>
         <color>ff000000</color>
         <hotSpot x="0.5"  y="0.5" xunits="fraction" yunits="fraction"/>
      </IconStyle>
      <LabelStyle>
         <color>00000000</color>
      </LabelStyle>
    </Style>

    <Style id="YellowDotIcon">
      <IconStyle>
         <scale>0.2</scale>
         <Icon>
            <href>dot.png</href>
         </Icon>
         <color>ff00ffff</color>
         <hotSpot x="0.5"  y="0.5" xunits="fraction" yunits="fraction"/>
      </IconStyle>
      <LabelStyle>
         <color>00000000</color>
      </LabelStyle>
    </Style>

    <Style id="RedXIcon">
      <IconStyle>
         <scale>0.2</scale>
         <color>ff0000ff</color>
         <Icon>
            <href>x.png</href>
         </Icon>
          <hotSpot x="0.5"  y="0.5" xunits="fraction" yunits="fraction"/>
      </IconStyle>
      <LabelStyle>
         <color>00000000</color>
      </LabelStyle>
    </Style>

    <Style id="BlackXIcon">
      <IconStyle>
         <scale>0.2</scale>
         <Icon>
            <href>x.png</href>
         </Icon>
         <color>ff000000</color>
         <hotSpot x="0.5"  y="0.5" xunits="fraction" yunits="fraction"/>
      </IconStyle>
      <LabelStyle>
         <color>00000000</color>
      </LabelStyle>
    </Style>

    <Style id="YellowXIcon">
      <IconStyle>
         <scale>0.2</scale>
         <Icon>
            <href>x.png</href>
         </Icon>
         <color>ff00ffff</color>
         <hotSpot x="0.5"  y="0.5" xunits="fraction" yunits="fraction"/>
      </IconStyle>
      <LabelStyle>
         <color>00000000</color>
      </LabelStyle>
    </Style>
"""


point_template = """             <Point>
                     <altitudeMode>relativeToGround</altitudeMode>
                     <coordinates>{:.6f},{:.6f},1.000000</coordinates>
             </Point>
"""


timestep_header_template = """<Folder>
  <name>{date_string}:{certain}</name>
"""

one_run_header = """    <Placemark>
      <name>{certain} {status} Splots </name>
      <styleUrl>{style}</styleUrl>
      <TimeSpan id="ID">
        <begin>{start_time}</begin>     <!-- kml:dateTime -->
        <end>{end_time}</end>         <!-- kml:dateTime -->
      </TimeSpan>
      <MultiGeometry>
"""
one_run_footer = """      </MultiGeometry>
    </Placemark>
"""
timestep_footer = """
</Folder>
"""


def build_one_timestep(floating_positions,
                       beached_positions,
                       start_time,
                       end_time,
                       uncertain,
                       ):

    data = {'certain': "Uncertainty" if uncertain else "Best Guess",
            'start_time': start_time,
            'end_time': end_time,
            'date_string': start_time,
            }
    kml = []
    kml.append(timestep_header_template.format(**data))

    for status, positions in [('Floating', floating_positions),
                              ('Beached', beached_positions)]:
        color = "Red" if uncertain else "Yellow"
        data['style'] = "#" + color + "DotIcon" if status == "Floating" else "#" + color + "XIcon"

        data['status'] = status
        kml.append(one_run_header.format(**data))

        for point in positions:
            kml.append(point_template.format(*point[:2]))
        kml.append(one_run_footer)
    kml.append(timestep_footer)

    return "".join(kml)

footer = """
  </Document>
</kml>
"""
# These icons (these are base64 encoded 3-pixel sized dots in a 32x32 transparent PNG)
#   these were encoded by the "build_icons" script
DOT = "iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAAJOgAACToB8GSSSgAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAAEASURBVFiF7ZY7DsIwEEQfET09Ej11lFtwK06Re3ANlCoFPQpnoGJoHClCXpOPg10wUhonnnlyvF5vJJFSRdL0P0AOANsZcwqgAkrg6MZuQANcgdckN0ljn52kWlInW537ZjfWd2z4SVIbCP5U6+ZEAThLek4I7/V0cxcBnGaGDyGCK/Htn09ZdkutAnsiBFBHCO9VWzkb+XtBAdyB/Ywy9ekBHPCUqHUQVRHDcV6V74UFUEYMD3paAEdjfIm8nsl7gQVwWyHL62kBNCsAeD2zLcMXcIkUjvPyt+nASZj8KE7ejLJox1lcSIZ7IvqVzCrDkKJeSucARFW2veAP8DO9AXV74Qmb/4vgAAAAAElFTkSuQmCC"
X = "iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAAN1wAADdcBQiibeAAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAAHKSURBVFiFrdXLq01hGMfx12HMzMCUU4zFQEYiROYkEkkpHbeTXI5LSDqHtomBEJGY+RMMGBlKKWVmaiDXzvExsN7admuv9azLU89k7ef5fb/ruhMSVuIy3uEOVhXH++w1mMEbnMFSpITl+Ob/+oOpHuFHiszh+oIVCbPGVx8Sh0vguaYT3lcIdJU4VAGHtwm3agTaShysgcMgYUNAoKnEgQAcVueFqR4l9mMhkHVJ8RbkPt6DxL4g/EreGQ3oIrE3CL86vFd2FidaSOzBfGDn+ihv3KU82UBidxB+o4xV9TBFJSKX/eY4Tt0TfSooUVWzVYzIO326A3yuLj/6YWkjcTuSHRVImG4AH0RzJ1K8PqSUFoKzn8KpQdNd+N3wFoT+OyLwnfjVEB6WqIPv6AAPSVTBt+NnR3itxDj4tiD8Hs52kSiDb8WPQOB9LCp2WkuMwrcE4Q8xMbJ7ro3EcMBmfA8EPCqBt5bIi5uC8McV8Nznm0gkLMPXwMKTADz3haDExoRjgcGnWByEN5EYJLyuGXrWAp57pib7Y8K1ioHnHeC5L1bkP0iYHPPjCyzpCK+SmMdkHliLl8XBVzjaIzz3Ov++H59xF+uR/gJmOo2+fdNArAAAAABJRU5ErkJggg=="




#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author:        David C. Morrill
#  Original Date: 06/21/2002
#
#  Rewritten as a C-based type extension: 06/21/2004
#
#------------------------------------------------------------------------------

"""
Defines the 'core' traits for the Traits package. A trait is a type definition
that can be used for normal Python object attributes, giving the attributes
some additional characteristics:

Initialization:
    Traits have predefined values that do not need to be explicitly
    initialized in the class constructor or elsewhere.
Validation:
    Trait attributes have flexible, type-checked values.
Delegation:
    Trait attributes' values can be delegated to other objects.
Notification:
    Trait attributes can automatically notify interested parties when
    their values change.
Visualization:
    Trait attributes can automatically construct (automatic or
    programmer-defined) user interfaces that allow their values to be
    edited or displayed)

.. note:: 'trait' is a synonym for 'property', but is used instead of the
    word 'property' to differentiate it from the Python language 'property'
    feature.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from __future__ import absolute_import

import sys
from types import FunctionType, MethodType
NoneType = type(None)   # Python 3's types does not include NoneType

from . import trait_handlers
from .ctraits import cTrait
from .trait_errors import TraitError
from .trait_base import (SequenceTypes, Self, Undefined, Missing, TypeTypes,
    add_article)

from .trait_handlers import (TraitHandler, TraitInstance, TraitFunction,
    TraitCoerceType, TraitCastType, TraitEnum, TraitCompound, TraitMap,
    TraitString, ThisClass, TraitType, _arg_count, _read_only, _write_only,
    _undefined_get, _undefined_set)


#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

# Mapping from 'ctrait' default value types to a string representation:
KindMap = {
   0: 'value',
   1: 'value',
   2: 'self',
   3: 'list',
   4: 'dict',
   5: 'list',
   6: 'dict',
   7: 'factory',
   8: 'method'
}

#-------------------------------------------------------------------------------
#  Editor factory functions:
#-------------------------------------------------------------------------------

PasswordEditor      = None
MultilineTextEditor = None
SourceCodeEditor    = None
HTMLTextEditor      = None
PythonShellEditor   = None
DateEditor          = None
TimeEditor          = None

def password_editor ( auto_set=True, enter_set=False ):
    """ Factory function that returns an editor for passwords.
    """
    global PasswordEditor

    if PasswordEditor is None:
        from traitsui.api import TextEditor
        PasswordEditor = TextEditor( password = True ,
                                     auto_set   = auto_set,
                                     enter_set  = enter_set )

    return PasswordEditor

def multi_line_text_editor ( auto_set=True, enter_set=False ):
    """ Factory function that returns a text editor for multi-line strings.
    """
    global MultilineTextEditor

    if MultilineTextEditor is None:
        from traitsui.api import TextEditor
        MultilineTextEditor = TextEditor( multi_line = True,
                                          auto_set   = auto_set,
                                          enter_set  = enter_set )

    return MultilineTextEditor

def code_editor ( ):
    """ Factory function that returns an editor that treats a multi-line string
    as source code.
    """
    global SourceCodeEditor

    if SourceCodeEditor is None:
        from traitsui.api import CodeEditor
        SourceCodeEditor = CodeEditor()

    return SourceCodeEditor

def html_editor ( ):
    """ Factory function for an "editor" that displays a multi-line string as
    interpreted HTML.
    """
    global HTMLTextEditor

    if HTMLTextEditor is None:
        from traitsui.api import HTMLEditor
        HTMLTextEditor = HTMLEditor()

    return HTMLTextEditor

def shell_editor ( ):
    """ Factory function that returns a Python shell for editing Python values.
    """
    global PythonShellEditor

    if PythonShellEditor is None:
        from traitsui.api import ShellEditor
        PythonShellEditor = ShellEditor()

    return PythonShellEditor

def time_editor ( ):
    """ Factory function that returns a Time editor for editing Time values.
    """
    global TimeEditor

    if TimeEditor is None:
        from traitsui.api import TimeEditor
        TimeEditor = TimeEditor()

    return TimeEditor

def date_editor ( ):
    """ Factory function that returns a Date editor for editing Date values.
    """
    global DateEditor

    if DateEditor is None:
        from traitsui.api import DateEditor
        DateEditor = DateEditor()

    return DateEditor

#-------------------------------------------------------------------------------
#  'CTrait' class (extends the underlying cTrait c-based type):
#-------------------------------------------------------------------------------

class CTrait ( cTrait ):
    """ Extends the underlying C-based cTrait type.
    """

    #---------------------------------------------------------------------------
    #  Allows a derivative trait to be defined from this one:
    #---------------------------------------------------------------------------

    def __call__ ( self, *args, **metadata ):
        handler = self.handler
        if isinstance( handler, TraitType ):
            dict = (self.__dict__ or {}).copy()
            dict.update( metadata )

            return handler( *args, **dict )

        metadata.setdefault( 'parent', self )
        return Trait( *(args + ( self, )), **metadata )

    #---------------------------------------------------------------------------
    #  (Python) property definitions:
    #---------------------------------------------------------------------------

    def __get_default ( self ):
        kind, value = self.default_value()
        if kind in ( 2, 7, 8 ):
            return Undefined

        if kind in ( 4, 6 ):
            return value.copy()

        if kind in ( 3, 5 ):
            return value[:]

        return value

    default = property( __get_default )

    def __get_default_kind ( self ):
        return KindMap[ self.default_value()[0] ]

    default_kind = property( __get_default_kind )

    def __get_trait_type ( self ):
        handler = self.handler
        if handler is not None:
            return handler
        else:
            from .trait_types import Any
            return Any

    trait_type = property( __get_trait_type )

    def __get_inner_traits ( self ):
        handler = self.handler
        if handler is not None:
            return handler.inner_traits()

        return ()

    inner_traits = property( __get_inner_traits )

    #---------------------------------------------------------------------------
    #  Returns whether or not this trait is of a specified trait type:
    #---------------------------------------------------------------------------

    def is_trait_type ( self, trait_type ):
        """ Returns whether or not this trait is of a specified trait type.
        """
        return isinstance( self.trait_type, trait_type )

    #---------------------------------------------------------------------------
    #  Returns the user interface editor associated with the trait:
    #---------------------------------------------------------------------------

    def get_editor ( self ):
        """ Returns the user interface editor associated with the trait.
        """
        from traitsui.api import EditorFactory

        # See if we have an editor:
        editor = self.editor
        if editor is None:

            # Else see if the trait handler has an editor:
            handler = self.handler
            if handler is not None:
                editor = handler.get_editor( self )

            # If not, give up and use a default text editor:
            if editor is None:
                from traitsui.api import TextEditor
                editor = TextEditor

        # If the result is not an EditorFactory:
        if not isinstance( editor, EditorFactory ):
            # Then it should be a factory for creating them:
            args   = ()
            traits = {}
            if type( editor ) in SequenceTypes:
                for item in editor[:]:
                    if type( item ) in SequenceTypes:
                        args = tuple( item )
                    elif isinstance( item, dict ):
                        traits = item
                        if traits.get( 'trait', 0 ) is None:
                            traits = traits.copy()
                            traits[ 'trait' ] = self
                    else:
                        editor = item
            editor = editor( *args, **traits )

        # Cache the result:
        self.editor = editor

        # Return the resulting EditorFactory object:
        return editor

    #---------------------------------------------------------------------------
    #  Returns the help text for a trait:
    #---------------------------------------------------------------------------

    def get_help ( self, full = True ):
        """ Returns the help text for a trait.

        Parameters
        ----------
        full : bool
            Indicates whether to return the value of the *help* attribute of
            the trait itself.

        Description
        -----------
        If *full* is False or the trait does not have a **help** string,
        the returned string is constructed from the **desc** attribute on the
        trait and the **info** string on the trait's handler.
        """
        if full:
            help = self.help
            if help is not None:
                return help

        handler = self.handler
        if handler is not None:
            info = 'must be %s.' % handler.info()
        else:
            info = 'may be any value.'

        desc = self.desc
        if self.desc is None:
            return info.capitalize()

        return 'Specifies %s and %s' % ( desc, info )

    #---------------------------------------------------------------------------
    #  Returns a description of the trait:
    #---------------------------------------------------------------------------

    def full_info ( self, object, name, value ):
        """ Returns a description of the trait.
        """
        handler = self.handler
        if handler is not None:
            return handler.full_info( object, name, value )

        return 'any value'

    #---------------------------------------------------------------------------
    #  Returns a description of the trait:
    #---------------------------------------------------------------------------

    def info ( self ):
        """ Returns a description of the trait.
        """
        handler = self.handler
        if handler is not None:
            return handler.info()

        return 'any value'

    #---------------------------------------------------------------------------
    #  Returns the pickleable form of a CTrait object:
    #---------------------------------------------------------------------------

    def __reduce_ex__ ( self, protocol ):
        return ( __newobj__, ( self.__class__, 0 ), self.__getstate__() )

    #---------------------------------------------------------------------------
    #  Registers listeners on an assigned 'TraitValue' object's 'value'
    #  property:
    #---------------------------------------------------------------------------

    def _register ( self, object, name ):
        """ Registers listeners on an assigned 'TraitValue' object's 'value'
            property.
        """
        def handler ( ):
            object.trait_property_changed( name, None )

        tv       = self._trait_value
        handlers = tv._handlers
        if handlers is None:
            tv._handlers = handlers = {}
        handlers[ ( id( object ), name ) ] = handler

        tv.on_trait_change( handler, 'value' )

    #---------------------------------------------------------------------------
    #  Unregisters listeners on an assigned 'TraitValue' object's 'value'
    #  property:
    #---------------------------------------------------------------------------

    def _unregister ( self, object, name ):
        """ Unregisters listeners on an assigned 'TraitValue' object's 'value'
            property.
        """
        tv       = self._trait_value
        handlers = tv._handlers
        key      = ( id( object ), name )
        handler  = handlers.get( key )
        if handler is not None:
            del handlers[ key ]
            tv.on_trait_change( handler, 'value', remove = True )

# Make sure the Python-level version of the trait class is known to all
# interested parties:
from . import ctraits
ctraits._ctrait( CTrait )

#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

ConstantTypes    = ( NoneType, int, long, float, complex, str, unicode )

PythonTypes      = ( str, unicode, int, long, float, complex, list, tuple,
                     dict, FunctionType, MethodType, type, NoneType )

if sys.version_info[0] < 3:
    from types import InstanceType,ClassType
    PythonTypes = PythonTypes[:-2] + (InstanceType,ClassType) + PythonTypes[2:]


CallableTypes    = ( FunctionType, MethodType )

TraitTypes       = ( TraitHandler, CTrait )

DefaultValues = {
    str:  '',
    unicode: u'',
    int:     0,
    long:    0L,
    float:   0.0,
    complex: 0j,
    list:    [],
    tuple:   (),
    dict:    {},
    bool: False
}

DefaultValueSpecial = [ Missing, Self ]
DefaultValueTypes   = [ list, dict ]

#-------------------------------------------------------------------------------
#  Function used to unpickle new-style objects:
#-------------------------------------------------------------------------------

def __newobj__ ( cls, *args ):
    """ Unpickles new-style objects.
    """
    return cls.__new__( cls, *args )

#-------------------------------------------------------------------------------
#  Returns the type of default value specified:
#-------------------------------------------------------------------------------

def _default_value_type ( default_value ):
    try:
        return DefaultValueSpecial.index( default_value ) + 1
    except:
        try:
            return DefaultValueTypes.index( type( default_value ) ) + 3
        except:
            return 0

#-------------------------------------------------------------------------------
#  'TraitFactory' class:
#-------------------------------------------------------------------------------

class TraitFactory ( object ):
    ### Need a docstring here.

    #---------------------------------------------------------------------------
    #  Initializes the object:
    #---------------------------------------------------------------------------

    def __init__ ( self, maker_function = None ):
        if maker_function is not None:
            self.maker_function = maker_function
            self.__doc__ = maker_function.__doc__

    #---------------------------------------------------------------------------
    #  Creates a CTrait instance:
    #---------------------------------------------------------------------------

    def __call__ ( self, *args, **metadata ):
        return self.maker_function( *args, **metadata )

class TraitImportError ( TraitFactory ):
    """ Defines a factory class for deferring import problems until encountering
        code that actually tries to use the unimportable trait.
    """

    #---------------------------------------------------------------------------
    #  Initializes the object:
    #---------------------------------------------------------------------------

    def __init__ ( self, message ):
        self.message = message

    #---------------------------------------------------------------------------
    #  Creates a CTrait instance:
    #---------------------------------------------------------------------------

    def __call__ ( self, *args, **metadata ):
        raise TraitError( self.message )

#-------------------------------------------------------------------------------
#  Returns a trait created from a TraitFactory instance:
#-------------------------------------------------------------------------------

_trait_factory_instances = {}

def trait_factory ( trait ):
    global _trait_factory_instances

    tid = id( trait )
    if tid not in _trait_factory_instances:
        _trait_factory_instances[ tid ] = trait()

    return _trait_factory_instances[ tid ]

#-------------------------------------------------------------------------------
#  Casts a CTrait or TraitFactory to a CTrait but returns None if it is neither:
#-------------------------------------------------------------------------------

def trait_cast ( something ):
    """ Casts a CTrait, TraitFactory or TraitType to a CTrait but returns None
        if it is none of those.
    """
    if isinstance( something, CTrait ):
        return something

    if isinstance( something, TraitFactory ):
        return trait_factory( something )

    if isinstance( something, type ) and issubclass( something, TraitType ):
        return something().as_ctrait()

    if isinstance( something, TraitType ):
        return something.as_ctrait()

    return None

#-------------------------------------------------------------------------------
#  Attempts to cast a value to a trait. Returns either a trait or the original
#  value:
#-------------------------------------------------------------------------------

def try_trait_cast ( something ):
    """ Attempts to cast a value to a trait. Returns either a trait or the
        original value.
    """
    return trait_cast( something ) or something

#-------------------------------------------------------------------------------
#  Returns a trait derived from its input:
#-------------------------------------------------------------------------------

def trait_from ( something ):
    """ Returns a trait derived from its input.
    """
    from .trait_types import Any

    if isinstance( something, CTrait ):
        return something

    if something is None:
        something = Any

    if isinstance( something, TraitFactory ):
        return trait_factory( something )

    if isinstance( something, type ) and issubclass( something, TraitType ):
        return something().as_ctrait()

    if isinstance( something, TraitType ):
        return something.as_ctrait()

    return Trait( something )

# Patch the reference to 'trait_from' in 'trait_handlers.py':
trait_handlers.trait_from = trait_from

#--- 'instance' traits ---------------------------------------------------------

class _InstanceArgs ( object ):

    def __init__ ( self, factory, args, kw ):
        self.args = ( factory, ) + args
        self.kw   = kw

#--- 'creates a run-time default value' ----------------------------------------

class Default ( object ):
    """ Generates a value the first time it is accessed.

    A Default object can be used anywhere a default trait value would normally
    be specified, to generate a default value dynamically.
    """
    def __init__ ( self, func = None, args = (), kw = None ):
        self.default_value = ( func, args, kw )

#-------------------------------------------------------------------------------
#  Factory function for creating C-based traits:
#-------------------------------------------------------------------------------

def Trait ( *value_type, **metadata ):
    """ Creates a trait definition.

    Parameters
    ----------
    This function accepts a variety of forms of parameter lists:

    +-------------------+---------------+-------------------------------------+
    | Format            | Example       | Description                         |
    +===================+===============+=====================================+
    | Trait(*default*)  | Trait(150.0)  | The type of the trait is inferred   |
    |                   |               | from the type of the default value, |
    |                   |               | which must be in *ConstantTypes*.   |
    +-------------------+---------------+-------------------------------------+
    | Trait(*default*,  | Trait(None,   | The trait accepts any of the        |
    | *other1*,         | 0, 1, 2,      | enumerated values, with the first   |
    | *other2*, ...)    | 'many')       | value being the default value. The  |
    |                   |               | values must be of types in          |
    |                   |               | *ConstantTypes*, but they need not  |
    |                   |               | be of the same type. The *default*  |
    |                   |               | value is not valid for assignment   |
    |                   |               | unless it is repeated later in the  |
    |                   |               | list.                               |
    +-------------------+---------------+-------------------------------------+
    | Trait([*default*, | Trait([None,  | Similar to the previous format, but |
    | *other1*,         | 0, 1, 2,      | takes an explicit list or a list    |
    | *other2*, ...])   | 'many'])      | variable.                           |
    +-------------------+---------------+-------------------------------------+
    | Trait(*type*)     | Trait(Int)    | The *type* parameter must be a name |
    |                   |               | of a Python type (see               |
    |                   |               | *PythonTypes*). Assigned values     |
    |                   |               | must be of exactly the specified    |
    |                   |               | type; no casting or coercion is     |
    |                   |               | performed. The default value is the |
    |                   |               | appropriate form of zero, False,    |
    |                   |               | or emtpy string, set or sequence.   |
    +-------------------+---------------+-------------------------------------+
    | Trait(*class*)    |::             | Values must be instances of *class* |
    |                   |               | or of a subclass of *class*. The    |
    |                   | class MyClass:| default value is None, but None     |
    |                   |    pass       | cannot be assigned as a value.      |
    |                   | foo = Trait(  |                                     |
    |                   | MyClass)      |                                     |
    +-------------------+---------------+-------------------------------------+
    | Trait(None,       |::             | Similar to the previous format, but |
    | *class*)          |               | None *can* be assigned as a value.  |
    |                   | class MyClass:|                                     |
    |                   |   pass        |                                     |
    |                   | foo = Trait(  |                                     |
    |                   | None, MyClass)|                                     |
    +-------------------+---------------+-------------------------------------+
    | Trait(*instance*) |::             | Values must be instances of the     |
    |                   |               | same class as *instance*, or of a   |
    |                   | class MyClass:| subclass of that class. The         |
    |                   |    pass       | specified instance is the default   |
    |                   | i = MyClass() | value.                              |
    |                   | foo =         |                                     |
    |                   |   Trait(i)    |                                     |
    +-------------------+---------------+-------------------------------------+
    | Trait(*handler*)  | Trait(        | Assignment to this trait is         |
    |                   | TraitEnum )   | validated by an object derived from |
    |                   |               | **traits.TraitHandler**.            |
    +-------------------+---------------+-------------------------------------+
    | Trait(*default*,  | Trait(0.0, 0.0| This is the most general form of    |
    | { *type* |        | 'stuff',      | the function. The notation:         |
    | *constant* |      | TupleType)    | ``{...|...|...}+`` means a list of  |
    | *dict* | *class* ||               | one or more of any of the items     |
    | *function* |      |               | listed between the braces. Thus, the|
    | *handler* |       |               | most general form of the function   |
    | *trait* }+ )      |               | consists of a default value,        |
    |                   |               | followed by one or more of several  |
    |                   |               | possible items. A trait defined by  |
    |                   |               | multiple items is called a          |
    |                   |               | "compound" trait.                   |
    +-------------------+---------------+-------------------------------------+

    All forms of the Trait function accept both predefined and arbitrary
    keyword arguments. The value of each keyword argument becomes bound to the
    resulting trait object as the value of an attribute having the same name
    as the keyword. This feature lets you associate metadata with a trait.

    The following predefined keywords are accepted:

    Keywords
    --------
    desc : str
        Describes the intended meaning of the trait. It is used in
        exception messages and fly-over help in user interfaces.
    label : str
        Provides a human-readable name for the trait. It is used to label user
        interface editors for traits.
    editor : traits.api.Editor
        Instance of a subclass Editor object to use when creating a user
        interface editor for the trait. See the "Traits UI User Guide" for
        more information on trait editors.
    comparison_mode : int
        Indicates when trait change notifications should be generated based upon
        the result of comparing the old and new values of a trait assignment:

        * 0 (NO_COMPARE): The values are not compared and a trait change
          notification is generated on each assignment.
        * 1 (OBJECT_IDENTITY_COMPARE): A trait change notification is
          generated if the old and new values are not the same object.
        * 2 (RICH_COMPARE): A trait change notification is generated if the
          old and new values are not equal using Python's
          'rich comparison' operator. This is the default.

    rich_compare : bool
        Indicates whether the basis for considering a trait attribute value to
        have changed is a "rich" comparison (True, the default), or simple
        object identity (False). This attribute can be useful in cases
        where a detailed comparison of two objects is very expensive, or where
        you do not care whether the details of an object change, as long as the
        same object is used.

            .. deprecated:: 3.0.3
                Use ``comparison_mode`` instead


    """
    return _TraitMaker( *value_type, **metadata ).as_ctrait()

#  Handle circular module dependencies:
trait_handlers.Trait = Trait

#-------------------------------------------------------------------------------
#  '_TraitMaker' class:
#-------------------------------------------------------------------------------

class _TraitMaker ( object ):

    # Ctrait type map for special trait types:
    type_map = {
       'event':    2,
       'constant': 7
    }

    #---------------------------------------------------------------------------
    #  Initialize the object:
    #---------------------------------------------------------------------------

    def __init__ ( self, *value_type, **metadata ):
        metadata.setdefault( 'type', 'trait' )
        self.define( *value_type, **metadata )

    #---------------------------------------------------------------------------
    #  Define the trait:
    #---------------------------------------------------------------------------

    def define ( self, *value_type, **metadata ):
        default_value_type = -1
        default_value      = handler = clone = None

        if len( value_type ) > 0:
            default_value = value_type[0]
            value_type    = value_type[1:]

            if ((len( value_type ) == 0) and
                (type( default_value ) in SequenceTypes)):
                default_value, value_type = default_value[0], default_value

            if len( value_type ) == 0:
                default_value = try_trait_cast( default_value )

                if default_value in PythonTypes:
                    handler       = TraitCoerceType( default_value )
                    default_value = DefaultValues.get( default_value )

                elif isinstance( default_value, CTrait ):
                    clone = default_value
                    default_value_type, default_value = clone.default_value()
                    metadata[ 'type' ] = clone.type

                elif isinstance( default_value, TraitHandler ):
                    handler       = default_value
                    default_value = None

                elif default_value is ThisClass:
                    handler       = ThisClass()
                    default_value = None

                else:
                    typeValue = type( default_value )

                    if isinstance(default_value, basestring):
                        string_options = self.extract( metadata, 'min_len',
                                                       'max_len', 'regex' )
                        if len( string_options ) == 0:
                            handler = TraitCastType( typeValue )
                        else:
                            handler = TraitString( **string_options )

                    elif typeValue in TypeTypes:
                        handler = TraitCastType( typeValue )

                    else:
                        metadata.setdefault( 'instance_handler',
                                             '_instance_changed_handler' )
                        handler = TraitInstance( default_value )
                        if default_value is handler.aClass:
                            default_value = DefaultValues.get( default_value )
            else:
                enum  = []
                other = []
                map   = {}
                self.do_list( value_type, enum, map, other )

                if (((len( enum )  == 1) and (enum[0] is None)) and
                    ((len( other ) == 1) and
                     isinstance( other[0], TraitInstance ))):
                    enum = []
                    other[0].allow_none()
                    metadata.setdefault( 'instance_handler',
                                         '_instance_changed_handler' )
                if len( enum ) > 0:
                    if (((len( map ) + len( other )) == 0) and
                        (default_value not in enum)):
                        enum.insert( 0, default_value )

                    other.append( TraitEnum( enum ) )

                if len( map ) > 0:
                    other.append( TraitMap( map ) )

                if len( other ) == 0:
                    handler = TraitHandler()

                elif len( other ) == 1:
                    handler = other[0]
                    if isinstance( handler, CTrait ):
                        clone, handler = handler, None
                        metadata[ 'type' ] = clone.type

                    elif isinstance( handler, TraitInstance ):
                        metadata.setdefault( 'instance_handler',
                                             '_instance_changed_handler' )

                        if default_value is None:
                            handler.allow_none()

                        elif isinstance( default_value, _InstanceArgs ):
                            default_value_type = 7
                            default_value = ( handler.create_default_value,
                                default_value.args, default_value.kw )

                        elif (len( enum ) == 0) and (len( map ) == 0):
                            aClass    = handler.aClass
                            typeValue = type( default_value )

                            if typeValue is dict:
                                default_value_type = 7
                                default_value = ( aClass, (), default_value )
                            elif not isinstance( default_value, aClass ):
                                if typeValue is not tuple:
                                    default_value = ( default_value, )
                                default_value_type = 7
                                default_value = ( aClass, default_value, None )
                else:
                    for i, item in enumerate( other ):
                        if isinstance( item, CTrait ):
                            if item.type != 'trait':
                                raise TraitError, ("Cannot create a complex "
                                    "trait containing %s trait." %
                                    add_article( item.type ) )
                            handler = item.handler
                            if handler is None:
                                break
                            other[i] = handler
                    else:
                        handler = TraitCompound( other )

        # Save the results:
        self.handler = handler
        self.clone   = clone

        if default_value_type < 0:
            if isinstance( default_value, Default ):
                default_value_type = 7
                default_value      = default_value.default_value
            else:
                if (handler is None) and (clone is not None):
                    handler = clone.handler

                if handler is not None:
                    default_value_type = handler.default_value_type
                    if default_value_type < 0:
                        try:
                            default_value = handler.validate( None, '',
                                                              default_value )
                        except:
                            pass

                if default_value_type < 0:
                    default_value_type = _default_value_type( default_value )

        self.default_value_type = default_value_type
        self.default_value      = default_value
        self.metadata           = metadata.copy()

    #---------------------------------------------------------------------------
    #  Determine the correct TraitHandler for each item in a list:
    #---------------------------------------------------------------------------

    def do_list ( self, list, enum, map, other ):
        for item in list:
            if item in PythonTypes:
                other.append( TraitCoerceType( item ) )
            else:
                item     = try_trait_cast( item )
                typeItem = type( item )

                if typeItem in ConstantTypes:
                    enum.append( item )

                elif typeItem in SequenceTypes:
                    self.do_list( item, enum, map, other )

                elif typeItem is dict:
                    map.update( item )

                elif typeItem in CallableTypes:
                    other.append( TraitFunction( item ) )

                elif item is ThisClass:
                    other.append( ThisClass() )

                elif isinstance( item, TraitTypes ):
                    other.append( item )

                else:
                    other.append( TraitInstance( item ) )

    #---------------------------------------------------------------------------
    #  Returns a properly initialized 'CTrait' instance:
    #---------------------------------------------------------------------------

    def as_ctrait ( self ):
        metadata = self.metadata
        trait    = CTrait( self.type_map.get( metadata.get( 'type' ), 0 ) )
        clone    = self.clone
        if clone is not None:
            trait.clone( clone )
            if clone.__dict__ is not None:
                trait.__dict__ = clone.__dict__.copy()

        trait.default_value( self.default_value_type, self.default_value )

        handler = self.handler
        if handler is not None:
            trait.handler = handler
            validate      = getattr( handler, 'fast_validate', None )
            if validate is None:
                validate = handler.validate

            trait.set_validate( validate )

            post_setattr = getattr( handler, 'post_setattr', None )
            if post_setattr is not None:
                trait.post_setattr = post_setattr
                trait.is_mapped( handler.is_mapped )

        # Note: The use of 'rich_compare' metadata is deprecated; use
        # 'comparison_mode' metadata instead:
        rich_compare = metadata.get( 'rich_compare' )
        if rich_compare is not None:
            trait.rich_comparison( rich_compare is True )

        comparison_mode = metadata.get( 'comparison_mode' )
        if comparison_mode is not None:
            trait.comparison_mode( comparison_mode )

        trait.value_allowed( metadata.get( 'trait_value', False ) is True )

        if len( metadata ) > 0:
            if trait.__dict__ is None:
                trait.__dict__ = metadata
            else:
                trait.__dict__.update( metadata )

        return trait

    #---------------------------------------------------------------------------
    #  Extract a set of keywords from a dictionary:
    #---------------------------------------------------------------------------

    def extract ( self, from_dict, *keys ):
        to_dict = {}
        for key in keys:
            if key in from_dict:
                to_dict[ key ] = from_dict[ key ]
                del from_dict[ key ]
        return to_dict

#-------------------------------------------------------------------------------
#  Factory function for creating C-based trait properties:
#-------------------------------------------------------------------------------

def Property ( fget = None, fset = None, fvalidate = None, force = False,
               handler = None, trait = None, **metadata ):
    """ Returns a trait whose value is a Python property.

    Parameters
    ----------
    fget : function
        The "getter" function for the property.
    fset : function
        The "setter" function for the property.
    fvalidate : function
        The validation function for the property. The method should return the
        value to set or raise TraitError if the new value is not valid.
    force : bool
        Indicates whether to use only the function definitions specified by
        **fget** and **fset**, and not look elsewhere on the class.
    handler : function
        A trait handler function for the trait.
    trait : Trait or value
        A trait definition or a value that can be converted to a trait that
        constrains the values of the property trait.

    Description
    -----------
    If no getter, setter or validate functions are specified (and **force** is
    not True), it is assumed that they are defined elsewhere on the class whose
    attribute this trait is assigned to. For example::

        class Bar(HasTraits):
            
            # A float traits Property that should be always positive.
            foo = Property(Float)
            
            # Shadow trait attribute
            _foo = Float

            def _set_foo(self,x):
                self._foo = x

            def _validate_foo(self, x):
                if x <= 0:
                    raise TraitError(
                        'foo property should be a positive number')
                return x

            def _get_foo(self):
                return self._foo

    You can use the **depends_on** metadata attribute to indicate that the
    property depends on the value of another trait. The value of **depends_on**
    is an extended name specifier for traits that the property depends on. The
    property will a trait change notification if any of the traits specified
    by **depends_on** change. For example::

        class Wheel ( Part ):
            axle     = Instanced( Axle )
            position = Property( depends_on = 'axle.chassis.position' )

    For details of the extended trait name syntax, refer to the on_trait_change()
    method of the HasTraits class.
    """
    metadata[ 'type' ] = 'property'

    # If no parameters specified, must be a forward reference (if not forced):
    if (not force) and (fset is None):
        sum = ((fget      is not None) +
               (fvalidate is not None) +
               (trait     is not None))
        if sum <= 1:
            if sum == 0:
                return ForwardProperty( metadata )

            handler = None
            if fget is not None:
                trait = fget

            if trait is not None:
                trait = trait_cast( trait )
                if trait is not None:
                    fvalidate = handler = trait.handler
                    if fvalidate is not None:
                        fvalidate = handler.validate

            if (fvalidate is not None) or (trait is not None):
                if 'editor' not in metadata:
                    if (trait is not None) and (trait.editor is not None):
                        metadata[ 'editor' ] = trait.editor

                return ForwardProperty( metadata, fvalidate, handler )

    if fget is None:
        metadata[ 'transient' ] = True
        if fset is None:
            fget = _undefined_get
            fset = _undefined_set
        else:
            fget = _write_only

    elif fset is None:
        fset = _read_only
        metadata[ 'transient' ] = True

    if trait is not None:
        trait   = trait_cast( trait )
        handler = trait.handler
        if (fvalidate is None) and (handler is not None):
            fvalidate = handler.validate

        if ('editor' not in metadata) and (trait.editor is not None):
            metadata[ 'editor' ] = trait.editor

    metadata.setdefault( 'depends_on', getattr( fget, 'depends_on', None ) )
    if ((metadata.get( 'depends_on' ) is not None) and
         getattr( fget, 'cached_property', False )):
        metadata.setdefault( 'cached', True )

    n     = 0
    trait = CTrait( 4 )
    trait.__dict__ = metadata.copy()
    if fvalidate is not None:
        n = _arg_count( fvalidate )

    trait.property( fget,      _arg_count( fget ),
                    fset,      _arg_count( fset ),
                    fvalidate, n )
    trait.handler = handler

    return trait

Property = TraitFactory( Property )

class ForwardProperty ( object ):
    """ Used to implement Property traits where accessor functions are defined
    implicitly on the class.
    """
    def __init__ ( self, metadata, validate = None, handler = None ):
        self.metadata = metadata.copy()
        self.validate = validate
        self.handler  = handler

#-------------------------------------------------------------------------------
#  Dictionary used to handle return type mapping special cases:
#-------------------------------------------------------------------------------

SpecialNames = {
###   'int':     trait_factory( Int ),
###   'long':    trait_factory( Long ),
###   'float':   trait_factory( Float ),
###   'complex': trait_factory( Complex ),
###   'str':     trait_factory( Str ),
###   'unicode': trait_factory( Unicode ),
###   'bool':    trait_factory( Bool ),
###   'list':    trait_factory( List ),
###   'tuple':   trait_factory( Tuple ),
###   'dict':    trait_factory( Dict )
}


#-- Date Trait definition ----------------------------------------------------
#Date = Instance(datetime.date, metadata = { 'editor': date_editor })


#-- Time Trait definition ----------------------------------------------------
#Time = Instance(datetime.time, metadata = { 'editor': time_editor })



#-------------------------------------------------------------------------------
#  Create predefined, reusable trait instances:
#-------------------------------------------------------------------------------

# Generic trait with 'object' behavior:
generic_trait = CTrait( 8 )

#-------------------------------------------------------------------------------
#  User interface related color and font traits:
#-------------------------------------------------------------------------------

def Color ( *args, **metadata ):
    """ Returns a trait whose value must be a GUI toolkit-specific color.

    Description
    -----------
    For wxPython, the returned trait accepts any of the following values:

    * A wx.Colour instance
    * A wx.ColourPtr instance
    * an integer whose hexadecimal form is 0x*RRGGBB*, where *RR* is the red
      value, *GG* is the green value, and *BB* is the blue value

    Default Value
    -------------
    For wxPython, 0x000000 (that is, white)
    """
    from traitsui.toolkit_traits import ColorTrait

    return ColorTrait( *args, **metadata )

Color = TraitFactory( Color )

def RGBColor ( *args, **metadata ):
    """ Returns a trait whose value must be a GUI toolkit-specific RGB-based
        color.

    Description
    -----------
    For wxPython, the returned trait accepts any of the following values:

    * A tuple of the form (*r*, *g*, *b*), in which *r*, *g*, and *b* represent
      red, green, and blue values, respectively, and are floats in the range
      from 0.0 to 1.0
    * An integer whose hexadecimal form is 0x*RRGGBB*, where *RR* is the red
      value, *GG* is the green value, and *BB* is the blue value

    Default Value
    -------------
    For wxPython, (0.0, 0.0, 0.0) (that is, white)
    """
    from traitsui.toolkit_traits import RGBColorTrait

    return RGBColorTrait( *args, **metadata )

RGBColor = TraitFactory( RGBColor )

def Font ( *args, **metadata ):
    """ Returns a trait whose value must be a GUI toolkit-specific font.

    Description
    -----------
    For wxPython, the returned trait accepts any of the following:

    * a wx.Font instance
    * a wx.FontPtr instance
    * a string describing the font, including one or more of the font family,
      size, weight, style, and typeface name.

    Default Value
    -------------
    For wxPython, 'Arial 10'
    """
    from traitsui.toolkit_traits import FontTrait

    return FontTrait( *args, **metadata )

Font = TraitFactory( Font )





from traitsui.api import View


class UpdateInfoView(View):
    """ A View for UpdateInfo objects that displays them in a nice HTML widget.
    Note that this view does NOT include any UI widgets for interacting with
    the user or prompting for a download; for that, use the UpdateCheckerDialog.
    """
    pass

